"""
.. deprecated:: 0.1

    This module is deprecated, and is only kept for reference purposes.
    Use `erlab.interactive.imagetool` instead.

"""

import sys
import time
from itertools import compress

import bottleneck as bn
import matplotlib.pyplot as plt
import numpy as np
import qtawesome as qta

# import numbagg
import xarray as xr
from joblib import Parallel, delayed
from matplotlib import colors
from matplotlib.backend_bases import _Mode
from matplotlib.backends.backend_qtagg import FigureCanvas, NavigationToolbar2QT
from matplotlib.backends.qt_compat import QtCore, QtGui, QtWidgets
from matplotlib.figure import Figure
from matplotlib.ticker import AutoLocator
from matplotlib.widgets import Widget

__all__ = ["mpl_itool"]


def qt_style_names():
    """Return a list of styles, default platform style first"""
    default_style_name = QtWidgets.QApplication.style().objectName().lower()
    result = []
    for style in QtWidgets.QStyleFactory.keys():
        if style.lower() == default_style_name:
            result.insert(0, style)
        else:
            result.append(style)
    return result


def change_style(style_name):
    QtWidgets.QApplication.setStyle(QtWidgets.QStyleFactory.create(style_name))


def colormap_to_QPixmap(name: str, h=64):
    """Convert matplotlib colormap to a 256-by-`h` QPixmap"""
    cmap = plt.colormaps[name]
    cmap_arr = cmap(np.tile(np.linspace(0, 1, 256), (h, 1))) * 255
    img = QtGui.QImage(
        cmap_arr.astype(np.uint8).data,
        cmap_arr.shape[1],
        cmap_arr.shape[0],
        QtGui.QImage.Format_RGBA8888,
    )
    return QtGui.QPixmap.fromImage(img)


def color_to_QColor(c, alpha=None):
    """Convert matplotlib color to QtGui.Qcolor."""
    return QtGui.QColor.fromRgbF(*colors.to_rgba(c, alpha=alpha))


def move_mean_centered(a, window, min_count=None, axis=-1):
    w = (window - 1) // 2
    shift = w + 1
    if min_count is None:
        min_count = w + 1
    pad_width = [(0, 0)] * a.ndim
    pad_width[axis] = (0, shift)
    a = np.pad(a, pad_width, constant_values=np.nan)
    val = bn.move_mean(a, window, min_count=min_count, axis=axis)
    return val[(slice(None),) * (axis % a.ndim) + (slice(w, -1),)]


def move_mean_centered_multiaxis(
    a,
    window_list,
    min_count_list=None,
    axis_list=[
        -1,
    ],
):
    if not isinstance(axis_list, list):
        axis_list = [
            axis_list,
        ]
    w_list = [(window - 1) // 2 for window in window_list]
    pad_width = [(0, 0)] * a.ndim
    slicer = [
        slice(None),
    ] * a.ndim
    if min_count_list is None:
        min_count_list = [w + 1 for w in w_list]
    for axis in axis_list:
        pad_width[axis] = (0, w_list[axis] + 1)
        slicer[axis] = slice(w_list[axis], -1)
    a = np.pad(a, pad_width, constant_values=np.nan)
    val = _move_mean_multiaxis_calc(a, window_list, min_count_list, axis_list)
    return val[tuple(slicer)]


def _move_mean_multiaxis_calc(a_padded, window_list, min_count_list, axis_list):
    val = a_padded
    for axis in axis_list:
        val = bn.move_mean(
            val, window_list[axis], min_count=min_count_list[axis], axis=axis
        )
    return val


def parse_data(data):
    if isinstance(data, xr.Dataset):
        try:
            data = data.spectrum
        except BaseException:
            raise TypeError(
                "input argument data must be a xarray.DataArray or a "
                "numpy.ndarray. Create an xarray.DataArray "
                "first, either with indexing on the Dataset or by "
                "invoking the `to_array()` method."
            ) from None
    elif isinstance(data, np.ndarray):
        data = xr.DataArray(data)
    return data


def is_vertical(artist):
    return artist.axes.get_xaxis_transform() == artist.get_transform()


def get_xy_x(a, b):
    return np.array([[a, 0.0], [a, 1.0], [b, 1.0], [b, 0.0], [a, 0.0]])


def get_xy_y(a, b):
    return np.array([[0.0, a], [0.0, b], [1.0, b], [1.0, a], [0.0, a]])


def get_true_indices(a):
    return list(compress(range(len(a)), a))


# https://www.pythonguis.com/widgets/qcolorbutton-a-color-selector-tool-for-pyqt/


class ColorButton(QtWidgets.QPushButton):
    """
    Custom Qt Widget to show a chosen color.

    Left-clicking the button shows the color-chooser, while
    right-clicking resets the color to None (no-color).
    """

    colorChanged = QtCore.Signal(object)

    def __init__(self, *args, color=None, **kwargs):
        super().__init__(*args, **kwargs)

        self._color = None
        self._default = color
        self.pressed.connect(self.onColorPicker)

        # Set the initial/default state.
        self.setColor(self._default)

    def setColor(self, color):
        self._color = color
        self.colorChanged.emit(color.getRgbF())
        if self._color:
            self.setStyleSheet(
                "QWidget { background-color: %s; border: 0; }"
                % self._color.name(QtGui.QColor.HexArgb)
            )
        else:
            self.setStyleSheet("")

    def color(self):
        return self._color

    def onColorPicker(self):
        """
        Show color-picker dialog to select color.

        Qt will use the native dialog by default.

        """
        dlg = QtWidgets.QColorDialog(self)
        if self._color:
            dlg.setCurrentColor(QtGui.QColor(self._color))
        if dlg.exec_():
            self.setColor(dlg.currentColor())

    def mousePressEvent(self, e):
        if e.button() == QtCore.Qt.RightButton:
            self.setColor(self._default)
        return super().mousePressEvent(e)


class mpl_itool(Widget):
    """A interactive tool based on `matplotlib` for exploring 3D data.

    For the tool to remain responsive you must
    keep a reference to it.

    Parameters
    ----------
    canvas : `matplotlib.backend_bases.FigureCanvasBase`
        The FigureCanvas that contains all the axes.
    axes : list of `matplotlib.axes.Axes`
        The `matplotlib.axes.Axes` to attach the cursor to. See Notes for the
        order of axes.
    data : `xarray.DataArray`
        The data to explore. Must have three coordinate axes.
    snap :  bool, default: True
        Snaps cursor to data pixels.
    parallel : bool, default: False
        Use multithreading. Currently has no performance improvement due
        to the python global interpreter lock.
    bench : bool, default: False
        Whether to print frames per second.

    Other Parameters
    ----------------
    **self.cursorprops
        `matplotlib.lines.Line2D` properties that control the appearance of the lines.
        See also `matplotlib.axes.Axes.axhline`.

    """

    # Notes
    # -----
    # Axes indices for 3D data:
    #     ┌───┬─────┐
    #     │ 1 │     │
    #     │───┤  3  │
    #     │ 4 │     │
    #     │───┼───┬─│
    #     │ 0 │ 5 │2│
    #     └───┴───┴─┘
    # Axes indices for 2D data:
    #     ┌───┬───┐
    #     │ 1 │   │
    #     │───┼───│
    #     │ 0 │ 2 │
    #     └───┴───┘

    def __init__(
        self,
        canvas,
        axes,
        data,
        snap=False,
        gamma=0.5,
        cmap="magma",
        useblit=True,
        parallel=False,
        bench=False,
        cursorprops={},
        lineprops={},
        fermilineprops={},
        **improps,
    ):
        self.canvas = canvas
        self.axes = axes
        self.data = parse_data(data)
        self.snap = snap
        self.gamma = gamma
        self.cmap = cmap
        self.useblit = useblit
        self.parallel = parallel
        self.bench = bench

        if self.useblit and not self.canvas.supports_blit:
            raise RuntimeError(
                "Canvas does not support blit. "
                "If running in ipython, add `%matplotlib qt`."
            )
        for ax in self.axes:
            for loc, spine in ax.spines.items():
                spine.set_position(("outward", 1))
        self.cursorprops = cursorprops
        self.lineprops = lineprops
        self.fermilineprops = fermilineprops
        self.improps = improps
        self.cursorprops.update(
            dict(
                linestyle="-",
                linewidth=0.8,
                color=colors.to_rgba(plt.rcParams.get("axes.edgecolor"), alpha=0.5),
                animated=self.useblit,
                visible=True,
            )
        )
        self.lineprops.update(
            dict(
                linestyle="-",
                linewidth=0.8,
                color=colors.to_rgba(plt.rcParams.get("axes.edgecolor"), alpha=1),
                animated=self.useblit,
                visible=True,
            )
        )
        self.fermilineprops.update(
            dict(
                linestyle="--",
                linewidth=0.8,
                color=colors.to_rgba(plt.rcParams.get("axes.edgecolor"), alpha=1),
                animated=False,
            )
        )
        self.improps.update(
            dict(
                animated=self.useblit,
                visible=True,
                interpolation="none",
                aspect="auto",
                origin="lower",
                norm=colors.PowerNorm(self.gamma),
                cmap=self.cmap,
                rasterized=True,
            )
        )
        self.spanprops = dict(
            # edgecolor=plt.rcParams.get('axes.edgecolor'),
            # lw=0.5, ls='--',
            facecolor=colors.to_rgba(self.cursorprops["color"], alpha=1),
            alpha=0.15,
            animated=self.useblit,
            visible=True,
        )
        self._get_middle_index = lambda x: len(x) // 2 - (1 if len(x) % 2 == 0 else 0)

        self.vals = self.data.values

        self.ndim = self.data.ndim
        self._assign_vals_T()
        self.avg_win = [
            1,
        ] * self.ndim
        self.clim_locked = False
        self.clim_list = [()] * self.ndim
        self.dims = self.data.dims
        self.coords = tuple(self.data[self.dims[i]] for i in range(self.ndim))
        self.shape = self.data.shape
        self.incs = tuple(
            self.coords[i][1] - self.coords[i][0] for i in range(self.ndim)
        )
        self.lims = tuple(
            (self.coords[i][0], self.coords[i][-1]) for i in range(self.ndim)
        )

        mids = tuple(self._get_middle_index(self.coords[i]) for i in range(self.ndim))
        self.cursor_pos = [self.coords[i][mids[i]] for i in range(self.ndim)]
        self._last_ind = list(mids)
        self._shift = False

        if self.ndim == 2:
            self.maps = (
                self.axes[0].imshow(
                    self.vals_T,
                    extent=(*self.lims[0], *self.lims[1]),
                    label="Main Image",
                    **self.improps,
                ),
            )
            self.hists = (
                self.axes[1].plot(
                    self.coords[0],
                    self.vals[:, mids[1]],
                    label="X Profile",
                    **self.lineprops,
                )[0],
                self.axes[2].plot(
                    self.vals[mids[0], :],
                    self.coords[1],
                    label="Y Profile",
                    **self.lineprops,
                )[0],
            )
            self.cursors = (
                self.axes[0].axvline(
                    self.coords[0][mids[0]], label="X Cursor", **self.cursorprops
                ),
                self.axes[1].axvline(
                    self.coords[0][mids[0]], label="X Cursor", **self.cursorprops
                ),
                self.axes[0].axhline(
                    self.coords[1][mids[1]], label="Y Cursor", **self.cursorprops
                ),
                self.axes[2].axhline(
                    self.coords[1][mids[1]], label="Y Cursor", **self.cursorprops
                ),
            )
            self.spans = (
                (
                    self.axes[0].axvspan(
                        self.coords[0][self._last_ind[0]],
                        self.coords[0][self._last_ind[0]],
                        label="X Span",
                        **self.spanprops,
                    ),
                    self.axes[1].axvspan(
                        self.coords[0][self._last_ind[0]],
                        self.coords[0][self._last_ind[0]],
                        label="X Span",
                        **self.spanprops,
                    ),
                ),
                (
                    self.axes[0].axhspan(
                        self.coords[1][self._last_ind[1]],
                        self.coords[1][self._last_ind[1]],
                        label="Y Span",
                        **self.spanprops,
                    ),
                    self.axes[2].axhspan(
                        self.coords[1][self._last_ind[1]],
                        self.coords[1][self._last_ind[1]],
                        label="Y Span",
                        **self.spanprops,
                    ),
                ),
            )
            self.scaling_axes = (self.axes[1].yaxis, self.axes[2].xaxis)
            self.ax_index = (0, 1, 2, 0, 1, 0, 2, 1, 2)
            self.span_ax_index = ((0, 1), (0, 2))
            self._only_axis = (
                (False, False, True, True, True, False, False),
                (False, True, False, False, False, True, True),
            )
        elif self.ndim == 3:
            self.maps = (
                self.axes[0].imshow(
                    self.vals_T[:, mids[2], :],
                    extent=(*self.lims[0], *self.lims[1]),
                    label="Main Image",
                    **self.improps,
                ),
                self.axes[4].imshow(
                    self.vals_T[mids[1], :, :],
                    extent=(*self.lims[0], *self.lims[2]),
                    label="Horiz Slice",
                    **self.improps,
                ),
                self.axes[5].imshow(
                    self.vals_T[:, :, mids[0]],
                    extent=(*self.lims[2], *self.lims[1]),
                    label="Vert Slice",
                    **self.improps,
                ),
            )
            self.hists = (
                self.axes[1].plot(
                    self.coords[0],
                    self.vals[:, mids[1], mids[2]],
                    label="X Profile",
                    **self.lineprops,
                )[0],
                self.axes[2].plot(
                    self.vals[mids[0], :, mids[2]],
                    self.coords[1],
                    label="Y Profile",
                    **self.lineprops,
                )[0],
                self.axes[3].plot(
                    self.coords[2],
                    self.vals[mids[0], mids[1], :],
                    label="Z Profile",
                    **self.lineprops,
                )[0],
            )
            self.cursors = (
                self.axes[0].axvline(
                    self.coords[0][mids[0]], label="X Cursor", **self.cursorprops
                ),
                self.axes[1].axvline(
                    self.coords[0][mids[0]], label="X Cursor", **self.cursorprops
                ),
                self.axes[4].axvline(
                    self.coords[0][mids[0]], label="X Cursor", **self.cursorprops
                ),
                self.axes[0].axhline(
                    self.coords[1][mids[1]], label="Y Cursor", **self.cursorprops
                ),
                self.axes[2].axhline(
                    self.coords[1][mids[1]], label="Y Cursor", **self.cursorprops
                ),
                self.axes[5].axhline(
                    self.coords[1][mids[1]], label="Y Cursor", **self.cursorprops
                ),
                self.axes[3].axvline(
                    self.coords[2][mids[2]], label="Z Cursor", **self.cursorprops
                ),
                self.axes[5].axvline(
                    self.coords[2][mids[2]], label="Z Cursor", **self.cursorprops
                ),
                self.axes[4].axhline(
                    self.coords[2][mids[2]], label="Z Cursor", **self.cursorprops
                ),
            )
            self.spans = (
                (
                    self.axes[0].axvspan(
                        self.coords[0][self._last_ind[0]],
                        self.coords[0][self._last_ind[0]],
                        label="X Span",
                        **self.spanprops,
                    ),
                    self.axes[1].axvspan(
                        self.coords[0][self._last_ind[0]],
                        self.coords[0][self._last_ind[0]],
                        label="X Span",
                        **self.spanprops,
                    ),
                    self.axes[4].axvspan(
                        self.coords[0][self._last_ind[0]],
                        self.coords[0][self._last_ind[0]],
                        label="X Span",
                        **self.spanprops,
                    ),
                ),
                (
                    self.axes[0].axhspan(
                        self.coords[1][self._last_ind[1]],
                        self.coords[1][self._last_ind[1]],
                        label="Y Span",
                        **self.spanprops,
                    ),
                    self.axes[2].axhspan(
                        self.coords[1][self._last_ind[1]],
                        self.coords[1][self._last_ind[1]],
                        label="Y Span",
                        **self.spanprops,
                    ),
                    self.axes[5].axhspan(
                        self.coords[1][self._last_ind[1]],
                        self.coords[1][self._last_ind[1]],
                        label="Y Span",
                        **self.spanprops,
                    ),
                ),
                (
                    self.axes[3].axvspan(
                        self.coords[2][self._last_ind[2]],
                        self.coords[2][self._last_ind[2]],
                        label="Z Span",
                        **self.spanprops,
                    ),
                    self.axes[5].axvspan(
                        self.coords[2][self._last_ind[2]],
                        self.coords[2][self._last_ind[2]],
                        label="Z Span",
                        **self.spanprops,
                    ),
                    self.axes[4].axhspan(
                        self.coords[2][self._last_ind[2]],
                        self.coords[2][self._last_ind[2]],
                        label="Z Span",
                        **self.spanprops,
                    ),
                ),
            )
            self.scaling_axes = (
                self.axes[1].yaxis,
                self.axes[2].xaxis,
                self.axes[3].yaxis,
            )
            if self.lims[-1][-1] * self.lims[-1][0] < 0:
                axes[3].axvline(0.0, label="Fermi Level", **self.fermilineprops)
            self.ax_index = (
                0,
                4,
                5,  # images
                1,
                2,
                3,  # profiles
                0,
                1,
                4,
                0,
                2,
                5,
                3,
                5,
                4,  # cursors
                1,
                2,
                3,
            )  # axes with dynamic limits
            self.span_ax_index = ((0, 1, 4), (0, 2, 5), (3, 5, 4))
            self._only_axis = (
                (
                    False,
                    False,
                    True,
                    False,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ),
                (
                    False,
                    True,
                    False,
                    True,
                    False,
                    True,
                    False,
                    False,
                    False,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                ),
                (
                    True,
                    False,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    True,
                    True,
                    True,
                ),
            )
            self.axes[3].set_xlabel(self.labelify(self.dims[2]))
            self.axes[4].set_ylabel(self.labelify(self.dims[2]))
            self.axes[5].set_xlabel(self.labelify(self.dims[2]))
            self.axes[3].xaxis.set_label_position("top")
            self.axes[3].set_xlim(self.lims[2])
            self.axes[4].set_ylim(self.lims[2])
            self.axes[5].set_xlim(self.lims[2])
            self.axes[3].set_yticks([])
            self.axes[3].ticklabel_format(
                axis="y", style="sci", scilimits=(-2, 3), useMathText=False
            )

        self.all = self.maps + self.hists + self.cursors

        self.axes[0].set_xlabel(self.labelify(self.dims[0]))
        self.axes[0].set_ylabel(self.labelify(self.dims[1]))
        self.axes[1].set_xlabel(self.labelify(self.dims[0]))
        self.axes[2].set_ylabel(self.labelify(self.dims[1]))
        self.axes[1].xaxis.set_label_position("top")
        self.axes[2].yaxis.set_label_position("right")
        self.axes[0].set_xlim(self.lims[0])
        self.axes[0].set_ylim(self.lims[1])
        self.axes[1].set_yticks([])
        self.axes[2].set_xticks([])
        self.axes[1].ticklabel_format(
            axis="y", style="sci", scilimits=(-2, 3), useMathText=False
        )
        self.axes[2].ticklabel_format(
            axis="x", style="sci", scilimits=(-2, 3), useMathText=False
        )

        if self.bench:
            self.counter = 0.0
            self.fps = 0.0
            self.lastupdate = time.time()

        if self.parallel:
            self.pool = Parallel(n_jobs=-1, require="sharedmem", verbose=0)

        self.averaged = [
            False,
        ] * self.ndim
        self.visible = True
        self.background = None
        self.needclear = False

        self.connect()
        # self.axes[1].xaxis.get_major_ticks()[0].label1.set_visible(False)
        # self.axes[3].xaxis.get_major_ticks()[-1].label1.set_visible(False)

    def connect(self):
        """Connect events."""
        self._cidmotion = self.canvas.mpl_connect("motion_notify_event", self.onmove)
        self._ciddraw = self.canvas.mpl_connect("draw_event", self.clear)
        self._cidpress = self.canvas.mpl_connect("key_press_event", self.onpress)
        self._cidrelease = self.canvas.mpl_connect("key_release_event", self.onrelease)

    def disconnect(self):
        """Disconnect events."""
        self.canvas.mpl_disconnect(self._cidmotion)
        self.canvas.mpl_disconnect(self._ciddraw)
        self.canvas.mpl_disconnect(self._cidpress)
        self.canvas.mpl_disconnect(self._cidrelease)

    def clear(self, event):
        """Clear objects and save background."""
        if self.ignore(event):
            return
        if self.useblit:
            self.background = self.canvas.copy_from_bbox(self.canvas.figure.bbox)
            for obj in self.all:
                obj.set_visible(False)
            for axis in range(self.ndim):
                for span in self.spans[axis]:
                    span.set_visible(False)
        for ax in self.scaling_axes:
            ax.set_ticks([])

    def labelify(self, dim):
        """Prettify some frequently used axis labels."""
        labelformats = dict(
            kx="$k_x$",
            ky="$k_y$",
            kz="$k_z$",
            alpha="$\\alpha$",
            beta="$\\beta$",
            theta="$\\theta$",
            phi="$\\phi$",
            chi="$\\chi$",
            eV="$E$",
        )
        try:
            return labelformats[dim]
        except KeyError:
            return dim

    def onpress(self, event):
        if event.key == "shift":
            self._shift = True

    def onrelease(self, event):
        if event.key == "shift":
            self._shift = False

    def _assign_vals_T(self):
        if self.ndim == 2:
            self.vals_T = self.vals.T
        elif self.ndim == 3:
            self.vals_T = np.transpose(self.vals, axes=(1, 2, 0))

    def set_cmap(self, cmap):
        self.cmap = cmap
        for im in self.maps:
            im.set_cmap(self.cmap)
        for obj in self.all:
            obj.set_visible(False)
        self._apply_change()

    def toggle_clim_lock(self, lock):
        if lock:
            self.clim_locked = True
            for i, m in enumerate(self.maps):
                self.clim_list[i] = m.get_clim()
        else:
            self.clim_locked = False

    def set_gamma(self, gamma):
        self.gamma = gamma
        self._apply_change()

    def set_index(self, axis, index):
        self._last_ind[axis] = index
        self.cursor_pos[axis] = self.coords[axis][index]
        self._apply_change(self._only_axis[axis])

    def set_value(self, axis, val):
        self._last_ind[axis] = self.get_index_of_value(axis, val)
        self.cursor_pos[axis] = val
        self._apply_change(self._only_axis[axis])

    def set_cursor_color(self, c):
        for cursor in self.cursors:
            cursor.set_color(c)
        self._apply_change()

    def set_line_color(self, c):
        for line in self.hists:
            line.set_color(c)
        self._apply_change()

    def set_navg(self, axis, n):
        self.avg_win[axis] = n
        if n == 1:
            self.averaged[axis] = False
            if not any(self.averaged):
                self.vals = self.data.values
                self._assign_vals_T()
                self._apply_change()
                return
        else:
            self.averaged[axis] = True

        if self.averaged[axis]:  # if already averaged
            self.vals = move_mean_centered_multiaxis(
                self.data.values,
                self.avg_win,
                axis_list=get_true_indices(self.averaged),
            )
        else:
            self.averaged[axis] = True
            self.vals = move_mean_centered(self.vals, window=n, axis=axis)
        self._assign_vals_T()
        self._apply_change()

    def update_spans(self):
        for axis in range(self.ndim):
            domain = (
                self.coords[axis][self._last_ind[axis]]
                - self.avg_win[axis] // 2 * self.incs[axis],
                self.coords[axis][self._last_ind[axis]]
                + (self.avg_win[axis] - 1) // 2 * self.incs[axis],
            )
            for span in self.spans[axis]:
                if is_vertical(span):
                    span.set_xy(get_xy_x(*domain))
                else:
                    span.set_xy(get_xy_y(*domain))
                span.set_visible(self.visible)
            if self.useblit:
                for i, span in list(zip(self.span_ax_index[axis], self.spans[axis])):
                    self.axes[i].draw_artist(span)

    def get_index_of_value(self, axis, val):
        # return np.rint((val-self.lims[axis][0])/self.incs[axis]).astype(int)
        return min(
            np.searchsorted(self.coords[axis] + 0.5 * self.incs[axis], val),
            self.shape[axis] - 1,
        )

    def onmove(self, event):
        if self.ignore(event):
            return
        if not event.button:
            if not self._shift:
                return
        if event.inaxes not in self.axes:
            return
        if not self.canvas.widgetlock.available(self):
            return
        self.needclear = True
        if not self.visible:
            return
        x, y, z = None, None, None
        if event.inaxes == self.axes[0]:
            dx, dy, dz = True, True, False
            x, y = event.xdata, event.ydata
        elif event.inaxes == self.axes[1]:
            dx, dy, dz = True, False, False
            x = event.xdata
        elif event.inaxes == self.axes[2]:
            dx, dy, dz = False, True, False
            y = event.ydata
        elif event.inaxes == self.axes[4]:
            dx, dy, dz = True, False, True
            x, z = event.xdata, event.ydata
        elif event.inaxes == self.axes[5]:
            dx, dy, dz = False, True, True
            z, y = event.xdata, event.ydata
        elif event.inaxes == self.axes[3]:
            dx, dy, dz = False, False, True
            z = event.xdata

        if self.ndim == 2:
            cond = (False, dy, dx, dx, dx, dy, dy)
        elif self.ndim == 3:
            cond = (
                dz,
                dy,
                dx,
                dy or dz,
                dx or dz,
                dx or dy,
                dx,
                dx,
                dx,
                dy,
                dy,
                dy,
                dz,
                dz,
                dz,
            )
        if dx:
            ind_x = self.get_index_of_value(0, x)
            if self.snap & (ind_x == self._last_ind[0]):
                dx = False
            else:
                self._last_ind[0] = ind_x
        if dy:
            ind_y = self.get_index_of_value(1, y)
            if self.snap & (ind_y == self._last_ind[1]):
                dy = False
            else:
                self._last_ind[1] = ind_y
        if dz:
            ind_z = self.get_index_of_value(2, z)
            if self.snap & (ind_z == self._last_ind[2]):
                dz = False
            else:
                self._last_ind[2] = ind_z

        if self.snap:
            self.cursor_pos = [
                self.coords[i][self._last_ind[i]] for i in range(self.ndim)
            ]
        else:
            self.cursor_pos = [x, y, z]
        self._apply_change(cond)
        if self.bench:
            self.print_time()

    def _apply_change(self, cond=None):
        if cond is None:
            cond = (True,) * len(self.all)
        if self.parallel:
            self.pool(delayed(self.set_data)(i) for i in get_true_indices(cond))
            self.pool(delayed(a.set_visible)(self.visible) for a in self.all)
            self._update()
        else:
            for i in get_true_indices(cond):
                self.set_data(i)
            for a in self.all:
                a.set_visible(self.visible)
            self._update()

    def _update(self):
        for ax in self.scaling_axes:
            ax.set_major_locator(AutoLocator())
            ax.axes.relim()
            ax.axes.autoscale_view()
        for i, m in enumerate(self.maps):
            m.set_norm(colors.PowerNorm(self.gamma))
            if self.clim_locked:
                m.set_clim(self.clim_list[i])

        if self.useblit:
            if self.background is not None:
                self.canvas.restore_region(self.background)
            if self.parallel:
                raise NotImplementedError
                # self.pool(delayed(self.axes[i].draw_artist)(art) for i, art in list(zip(
                #     (0, 4, 5, 1, 2, 3, 1, 2, 3),
                #     self.maps + self.hists + (self.axes[1].yaxis,
                #                               self.axes[2].xaxis,
                #                               self.axes[3].yaxis))))
                # self.pool(delayed(self.axes[i].draw_artist)(art) for i, art in list(zip(
                #     (0, 1, 4, 0, 2, 5, 3, 5, 4), self.cursors)))
            else:
                for i, art in list(zip(self.ax_index, self.all + self.scaling_axes)):
                    self.axes[i].draw_artist(art)
            if any(self.averaged):
                self.update_spans()
            self.canvas.blit()
        else:
            if any(self.averaged):
                self.update_spans()
            self.canvas.draw_idle()

    def print_time(self):
        now = time.time()
        dt = now - self.lastupdate
        if dt <= 0:
            dt = 0.000000000001
        fps2 = 1.0 / dt
        self.lastupdate = now
        self.fps = self.fps * 0.9 + fps2 * 0.1
        tx = f"Mean Frame Rate:  {self.fps:.3f} FPS"
        print(tx, end="\r")

    def set_data(self, i):
        if self.ndim == 2:
            self.set_data_2d(i)
        elif self.ndim == 3:
            self.set_data_3d(i)

    def set_data_2d(self, i):
        if i == 0:
            self.all[i].set_data(self.vals_T)
        elif i == 1:
            self.all[i].set_ydata(self.vals_T[self._last_ind[1], :])
        elif i == 2:
            self.all[i].set_xdata(self.vals_T[:, self._last_ind[0]])
        elif i in [3, 4]:
            self.all[i].set_xdata((self.cursor_pos[0], self.cursor_pos[0]))
        elif i in [5, 6]:
            self.all[i].set_ydata((self.cursor_pos[1], self.cursor_pos[1]))

    def set_data_3d(self, i):
        if i == 0:
            self.all[i].set_data(self.vals_T[:, self._last_ind[2], :])
        elif i == 1:
            self.all[i].set_data(self.vals_T[self._last_ind[1], :, :])
        elif i == 2:
            self.all[i].set_data(self.vals_T[:, :, self._last_ind[0]])
        elif i == 3:
            self.all[i].set_ydata(self.vals[:, self._last_ind[1], self._last_ind[2]])
        elif i == 4:
            self.all[i].set_xdata(self.vals[self._last_ind[0], :, self._last_ind[2]])
        elif i == 5:
            self.all[i].set_ydata(self.vals[self._last_ind[0], self._last_ind[1], :])
        elif i in [6, 7, 8]:
            self.all[i].set_xdata((self.cursor_pos[0], self.cursor_pos[0]))
        elif i in [9, 10, 11]:
            self.all[i].set_ydata((self.cursor_pos[1], self.cursor_pos[1]))
        elif i in [12, 13]:
            self.all[i].set_xdata((self.cursor_pos[2], self.cursor_pos[2]))
        elif i == 14:
            self.all[i].set_ydata((self.cursor_pos[2], self.cursor_pos[2]))

    def _drawpath(self):
        # ld = LineDrawer(self.canvas, self.axes[0])
        # points = ld.draw_line()
        # print(points)
        # TODO
        pass

    def _onselectpath(self, verts):
        print(verts)


class ImageToolNavBar(NavigationToolbar2QT):
    def __init__(self, canvas, parent, coordinates=True):
        self.parent = parent
        NavigationToolbar2QT.__init__(self, canvas, parent, coordinates=coordinates)

    def _icon(self, name):
        """
        Construct a `.QIcon` from an image file *name*, including the extension
        and relative to Matplotlib's "images" data directory.
        """
        name = name.replace(".png", "")
        icons_dict = dict(
            # back = qta.icon('ph.arrow-arc-left-fill'),
            # forward = qta.icon('ph.arrow-arc-right-fill'),
            # filesave = qta.icon('ph.floppy-disk-fill'),
            # home = qta.icon('ph.corners-out-fill'),
            # move = qta.icon('ph.arrows-out-cardinal-fill'),
            # qt4_editor_options = qta.icon('ph.palette-fill'),
            # zoom_to_rect = qta.icon('ph.crop-fill'),
            # subplots = qta.icon('ph.squares-four-fill'),
            back=qta.icon("msc.chevron-left"),
            forward=qta.icon("msc.chevron-right"),
            filesave=qta.icon("msc.save"),
            home=qta.icon("msc.debug-step-back"),
            move=qta.icon("msc.move"),
            qt4_editor_options=qta.icon("msc.graph-line"),
            zoom_to_rect=qta.icon("msc.search"),
            subplots=qta.icon("msc.editor-layout"),
        )
        try:
            return icons_dict[name]
        except BaseException:
            print(name)
            raise Exception
        # name = name.replace('.png', '_large.png')
        # pm = QtGui.QPixmap(str(cbook._get_data_path('images', name)))
        # _setDevicePixelRatio(pm, _devicePixelRatioF(self))
        # if self.palette().color(self.backgroundRole()).value() < 128:
        # icon_color = self.palette().color(self.foregroundRole())
        # mask = pm.createMaskFromColor(
        # QtGui.QColor('black'),
        # _enum("QtCore.Qt.MaskMode").MaskOutColor)
        # pm.fill(icon_color)
        # pm.setMask(mask)
        # return QtGui.QIcon(pm)


class ImageToolColors(QtWidgets.QDialog):
    def __init__(self, parent):
        self.parent = parent
        super().__init__(self.parent)
        self.setWindowTitle("Colors")

        self.cursor_default = color_to_QColor(self.parent.itool.cursorprops["color"])
        self.line_default = color_to_QColor(self.parent.itool.lineprops["color"])
        self.cursor_current = color_to_QColor(self.parent.itool.cursors[0].get_color())
        self.line_current = color_to_QColor(self.parent.itool.hists[0].get_color())

        if (self.cursor_default.getRgbF() == self.cursor_current.getRgbF()) & (
            self.line_default.getRgbF() == self.line_current.getRgbF()
        ):
            buttons = QtWidgets.QDialogButtonBox(
                QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
            )
        else:
            buttons = QtWidgets.QDialogButtonBox(
                QtWidgets.QDialogButtonBox.RestoreDefaults
                | QtWidgets.QDialogButtonBox.Ok
                | QtWidgets.QDialogButtonBox.Cancel
            )
            buttons.button(QtWidgets.QDialogButtonBox.RestoreDefaults).clicked.connect(
                self.reset_colors
            )
        buttons.rejected.connect(self.reject)
        buttons.accepted.connect(self.accept)

        cursorlabel = QtWidgets.QLabel("Cursors:")
        linelabel = QtWidgets.QLabel("Lines:")
        self.cursorpicker = ColorButton(color=self.cursor_current)
        self.cursorpicker.colorChanged.connect(self.parent.itool.set_cursor_color)
        self.linepicker = ColorButton(color=self.line_current)
        self.linepicker.colorChanged.connect(self.parent.itool.set_line_color)

        layout = QtWidgets.QGridLayout()
        layout.addWidget(cursorlabel, 0, 0)
        layout.addWidget(self.cursorpicker, 0, 1)
        layout.addWidget(linelabel, 1, 0)
        layout.addWidget(self.linepicker, 1, 1)
        layout.addWidget(buttons)
        self.setLayout(layout)

    def reject(self):
        self.cursorpicker.setColor(self.cursor_current)
        self.linepicker.setColor(self.line_current)
        super().reject()

    def reset_colors(self):
        self.cursorpicker.setColor(self.cursor_default)
        self.linepicker.setColor(self.line_default)


class ImageTool(QtWidgets.QMainWindow):
    def __init__(self, data, *args, **kwargs):
        super().__init__()
        self._main = QtWidgets.QWidget()
        self.setCentralWidget(self._main)
        self.layout = QtWidgets.QVBoxLayout(self._main)
        self.figure = Figure(
            figsize=(10, 10), dpi=70, frameon=True, layout="constrained"
        )
        self.data = parse_data(data)
        self.ndim = self.data.ndim
        if self.ndim == 3:
            gs = self.figure.add_gridspec(
                3, 3, width_ratios=(6, 4, 2), height_ratios=(2, 4, 6)
            )
            self._main_axes = self.figure.add_subplot(gs[2, 0])
            self.axes = [
                self._main_axes,
                self.figure.add_subplot(gs[0, 0], sharex=self._main_axes),
                self.figure.add_subplot(gs[2, 2], sharey=self._main_axes),
                self.figure.add_subplot(gs[:2, 1:]),
                self.figure.add_subplot(gs[1, 0], sharex=self._main_axes),
                self.figure.add_subplot(gs[2, 1], sharey=self._main_axes),
            ]
            self.axes[3].set_label("Z Profile Axes")
            self.axes[4].set_label("Horiz Slice Axes")
            self.axes[5].set_label("Vert Slice Axes")
            self.axes[4].label_outer()
            self.axes[5].label_outer()
            self.axes[3].xaxis.tick_top()
            self.axes[3].yaxis.tick_right()
        elif self.ndim == 2:
            gs = self.figure.add_gridspec(
                2, 2, width_ratios=(6, 2), height_ratios=(2, 6)
            )
            self._main_axes = self.figure.add_subplot(gs[1, 0])
            self.axes = [
                self._main_axes,
                self.figure.add_subplot(gs[0, 0], sharex=self._main_axes),
                self.figure.add_subplot(gs[1, 1], sharey=self._main_axes),
            ]
        self.axes[0].set_label("Main Image Axes")
        self.axes[1].set_label("X Profile Axes")
        self.axes[2].set_label("Y Profile Axes")
        self.axes[1].xaxis.tick_top()
        self.axes[2].yaxis.tick_right()

        self.main_canvas = FigureCanvas(self.figure)
        self.itool = mpl_itool(self.main_canvas, self.axes, self.data, *args, **kwargs)
        self.NavBar = ImageToolNavBar
        home_old = self.NavBar.home

        def home_new(self, *args):
            home_old(self, *args)
            axes = self.canvas.figure.axes
            axes[1].set_ylim(auto=True)
            axes[2].set_xlim(auto=True)
            if self.parent.itool.ndim == 3:
                axes[3].set_ylim(auto=True)

        pan_old = self.NavBar.pan

        def pan_new(self, *args):
            if self.mode == _Mode.PAN:
                self.parent.itool.connect()
            else:
                self.parent.itool.disconnect()
            pan_old(self, *args)

        zoom_old = self.NavBar.zoom

        def zoom_new(self, *args):
            if self.mode == _Mode.ZOOM:
                self.parent.itool.connect()
            else:
                self.parent.itool.disconnect()
            zoom_old(self, *args)

        self.NavBar.home = home_new
        self.NavBar.pan = pan_new
        self.NavBar.zoom = zoom_new
        self.addToolBar(
            QtCore.Qt.BottomToolBarArea, self.NavBar(self.main_canvas, self)
        )

        self.icons = dict(
            swap=qta.icon("msc.arrow-swap"),
            lock=qta.icon("msc.lock"),
            unlock=qta.icon("msc.unlock"),
        )

        self.cursortab = QtWidgets.QWidget()
        cursortab_content = QtWidgets.QHBoxLayout(self.cursortab)
        spinlabels = tuple(
            QtWidgets.QLabel(self.itool.dims[i]) for i in range(self.ndim)
        )
        self._cursor_spin = tuple(
            QtWidgets.QSpinBox(self.cursortab) for i in range(self.ndim)
        )
        self._cursor_dblspin = tuple(
            QtWidgets.QDoubleSpinBox(self.cursortab) for i in range(self.ndim)
        )
        for i in range(self.ndim):
            self._cursor_spin[i].setRange(0, self.itool.shape[i] - 1)
            self._cursor_spin[i].setSingleStep(1)
            self._cursor_spin[i].setValue(self.itool._last_ind[i])
            self._cursor_spin[i].setWrapping(True)
            self._cursor_spin[i].valueChanged.connect(
                lambda v, axis=i: self._cursor_index_changed(axis, v)
            )
            self._cursor_dblspin[i].setRange(*self.itool.lims[i])
            self._cursor_dblspin[i].setSingleStep(self.itool.incs[i])
            self._cursor_dblspin[i].setDecimals(3)
            self._cursor_dblspin[i].setValue(
                self.itool.coords[i][self.itool._last_ind[i]]
            )
            self._cursor_dblspin[i].setWrapping(True)
            self._cursor_dblspin[i].valueChanged.connect(
                lambda v, axis=i: self._cursor_value_changed(axis, v)
            )

        snap_check = QtWidgets.QCheckBox(self.cursortab)
        snap_check.setChecked(self.itool.snap)
        snap_check.stateChanged.connect(self._assign_snap)
        snap_label = QtWidgets.QLabel("Snap to data")
        snap_label.setBuddy(snap_check)

        for i in range(self.ndim):
            cursortab_content.addWidget(spinlabels[i])
            cursortab_content.addWidget(self._cursor_dblspin[i])
            cursortab_content.addWidget(self._cursor_spin[i])
            cursortab_content.addSpacing(20)
        cursortab_content.addStretch()
        cursortab_content.addWidget(snap_check)
        cursortab_content.addWidget(snap_label)

        self.colorstab = QtWidgets.QWidget()
        colorstab_content = QtWidgets.QHBoxLayout(self.colorstab)
        gamma_spin = QtWidgets.QDoubleSpinBox()
        gamma_spin.setToolTip("Colormap Gamma")
        gamma_spin.setSingleStep(0.01)
        gamma_spin.setRange(0.01, 100.0)
        gamma_spin.setValue(self.itool.gamma)
        gamma_spin.valueChanged.connect(self.itool.set_gamma)
        gamma_label = QtWidgets.QLabel("g")
        gamma_label.setBuddy(gamma_spin)
        self._cmap_combo = QtWidgets.QComboBox(self.colorstab)
        self._cmap_combo.setToolTip("Colormap")
        for name in plt.colormaps():
            self._cmap_combo.addItem(QtGui.QIcon(colormap_to_QPixmap(name)), name)
        self._cmap_combo.setCurrentText(self.itool.cmap)
        self._cmap_combo.setIconSize(QtCore.QSize(60, 15))
        self._cmap_combo.currentTextChanged.connect(self.itool.set_cmap)
        cmap_r_button = QtWidgets.QPushButton()
        cmap_r_button.setIcon(self.icons["swap"])
        cmap_r_button.setCheckable(True)
        cmap_r_button.toggled.connect(self.reverse_cmap)
        self._lock_check = QtWidgets.QPushButton(self.colorstab)
        self._lock_check.setIcon(self.icons["unlock"])
        self._lock_check.setCheckable(True)
        self._lock_check.toggled.connect(self._toggle_clim_lock)
        colors_button = QtWidgets.QPushButton("Colors")
        colors_button.clicked.connect(self._color_button_clicked)
        style_combo = QtWidgets.QComboBox(self.colorstab)
        style_combo.setToolTip("Qt Style")
        style_combo.addItems(qt_style_names())
        style_combo.textActivated.connect(change_style)
        style_label = QtWidgets.QLabel("Style:")
        style_label.setBuddy(style_combo)
        style_combo.setCurrentIndex(style_combo.findText("Fusion"))
        self.main_canvas.draw()
        colorstab_content.addWidget(gamma_label)
        colorstab_content.addWidget(gamma_spin)
        colorstab_content.addWidget(self._cmap_combo)
        colorstab_content.addWidget(cmap_r_button)
        colorstab_content.addWidget(self._lock_check)
        colorstab_content.addStretch()
        colorstab_content.addWidget(colors_button)
        colorstab_content.addStretch()
        colorstab_content.addWidget(style_label)
        colorstab_content.addWidget(style_combo)

        self.smoothtab = QtWidgets.QWidget()
        smoothtab_content = QtWidgets.QHBoxLayout(self.smoothtab)
        navg_label = tuple(
            QtWidgets.QLabel(self.itool.dims[i]) for i in range(self.ndim)
        )
        self._navg_spin = tuple(
            QtWidgets.QSpinBox(self.smoothtab) for i in range(self.ndim)
        )
        navg_resetbutton = QtWidgets.QPushButton("Reset")
        navg_resetbutton.clicked.connect(self._navg_reset)
        for i in range(self.ndim):
            self._navg_spin[i].setRange(1, self.itool.shape[i] - 1)
            self._navg_spin[i].setSingleStep(2)
            self._navg_spin[i].setValue(1)
            self._navg_spin[i].setWrapping(False)
            self._navg_spin[i].valueChanged.connect(
                lambda n, axis=i: self._navg_changed(axis, n)
            )
        for i in range(self.ndim):
            smoothtab_content.addWidget(navg_label[i])
            smoothtab_content.addWidget(self._navg_spin[i])
            smoothtab_content.addSpacing(20)
        smoothtab_content.addWidget(navg_resetbutton)
        smoothtab_content.addStretch()

        # self.pathtab = QtWidgets.QWidget()
        # pathtabcontent = QtWidgets.QHBoxLayout()
        # pathlabel = QtWidgets.QLabel('Add point: `space`\nRemove point: `delete`\nFinish selection: `enter`')
        # pathstart = QtWidgets.QPushButton()
        # pathstart.clicked.connect(self.itool._drawpath)
        # pathtabcontent.addWidget(pathlabel)
        # pathtabcontent.addWidget(pathstart)
        # self.pathtab.setLayout(pathtabcontent)

        self.tabwidget = QtWidgets.QTabWidget()
        self.tabwidget.addTab(self.cursortab, "Cursor")
        self.tabwidget.addTab(self.colorstab, "Appearance")
        self.tabwidget.addTab(self.smoothtab, "Smoothing")
        # self.tabwidget.addTab(self.pathtab, 'Path')

        self.layout.addWidget(self.main_canvas)
        self.layout.addWidget(self.tabwidget)
        self.main_canvas.mpl_connect("motion_notify_event", self.onmove_super)
        self.main_canvas.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.main_canvas.setFocus()
        self.main_canvas.draw()

    def onmove_super(self, event):
        if event.inaxes not in self.axes:
            return
        if not event.button:
            if not self.itool._shift:
                return
        for i in range(self.ndim):
            self._cursor_spin[i].blockSignals(True)
            self._cursor_spin[i].setValue(self.itool._last_ind[i])
            self._cursor_spin[i].blockSignals(False)
            self._cursor_dblspin[i].blockSignals(True)
            self._cursor_dblspin[i].setValue(
                self.itool.coords[i][self.itool._last_ind[i]]
            )
            self._cursor_dblspin[i].blockSignals(False)

    def reverse_cmap(self, v):
        if v:
            self.itool.set_cmap(plt.get_cmap(self._cmap_combo.currentText()).reversed())
        else:
            self.itool.set_cmap(plt.get_cmap(self._cmap_combo.currentText()))

    def _toggle_clim_lock(self, v):
        if v:
            self._lock_check.setIcon(self.icons["lock"])
        else:
            self._lock_check.setIcon(self.icons["unlock"])
        self.itool.toggle_clim_lock(v)

    def _navg_changed(self, axis, n):
        self.itool.set_navg(axis, n)

    def _navg_reset(self):
        for i in range(self.ndim):
            self._navg_spin[i].setValue(1)

    def _cursor_index_changed(self, axis, index):
        self._cursor_dblspin[axis].blockSignals(True)
        self.itool.set_index(axis, index)
        self._cursor_dblspin[axis].setValue(self.itool.coords[axis][index])
        self._cursor_dblspin[axis].blockSignals(False)

    def _cursor_value_changed(self, axis, value):
        self._cursor_spin[axis].blockSignals(True)
        self.itool.set_value(axis, value)
        self._cursor_spin[axis].setValue(self.itool._last_ind[axis])
        self._cursor_spin[axis].blockSignals(False)

    def _color_button_clicked(self, s):
        # print("click", s)
        dialog = ImageToolColors(self)
        if dialog.exec():
            # print("Success!")
            pass
        else:
            pass
            # print("Cancel!")

    def _assign_snap(self, value):
        self.itool.snap = value


def itoolmpl(data, *args, **kwargs):
    qapp = QtWidgets.QApplication.instance()
    if not qapp:
        qapp = QtWidgets.QApplication(sys.argv)
    # print(qapp.devicePixelRatio())
    mpl_style = "default"
    with plt.rc_context(
        {
            "text.usetex": False,
            # 'font.family':'SF Pro',
            # 'font.size':8,
            # 'font.stretch':'condensed',
            # 'mathtext.fontset':'cm',
            # 'font.family':'fantasy',
        }
    ):
        with plt.style.context(mpl_style):
            app = ImageTool(data, *args, **kwargs)
    change_style("Fusion")
    app.show()
    app.activateWindow()
    app.raise_()
    qapp.exec()



"""Various functions used in fitting.

.. currentmodule:: erlab.analysis.fit.functions

Modules
=======

.. autosummary::
   :toctree:

   dynamic
   general

"""

__all__ = [
    "TINY",
    "do_convolve",
    "do_convolve_y",
    "gaussian_wh",
    "lorentzian_wh",
    "fermi_dirac",
    "fermi_dirac_linbkg",
    "fermi_dirac_linbkg_broad",
    "step_linbkg_broad",
    "step_broad",
    "PolyFunc",
    "MultiPeakFunction",
    "FermiEdge2dFunc",
]

from erlab.analysis.fit.functions.general import (
    TINY,
    do_convolve,
    do_convolve_y,
    gaussian_wh,
    lorentzian_wh,
    fermi_dirac,
    fermi_dirac_linbkg,
    fermi_dirac_linbkg_broad,
    step_linbkg_broad,
    step_broad,
)

from erlab.analysis.fit.functions.dynamic import (
    PolyFunc,
    MultiPeakFunction,
    FermiEdge2dFunc,
)



"""Read & write ARPES data.

.. currentmodule:: erlab.io

This module provides functions that enables loading various files such as hdf5 files,
igor pro files, and ARPES data from different beamlines and laboratories.

Modules
=======

.. autosummary::
   :toctree: generated

   plugins
   dataloader
   utilities
   igor


For a single session, it is very common to use only one type of loader for a single
folder with all your data. Hence, the module provides a way to set a default loader for
a session. This is done using the :func:`set_loader` function. The same can be done for
the data directory using the :func:`set_data_dir` function.

For instructions on how to write a custom loader, see
:doc:`/generated/erlab.io.dataloader`.


Examples
--------

- View all registered loaders:

  >>> erlab.io.loaders

- Load data by explicitly specifying the loader:

  >>> dat = erlab.io.loaders["merlin"].load(...)


"""

__all__ = [
    "loaders",
    "set_loader",
    "loader_context",
    "set_data_dir",
    "load",
    "summarize",
    "load_experiment",
    "load_wave",
    "load_live",
    "open_hdf5",
    "load_hdf5",
    "save_as_hdf5",
    "save_as_netcdf",
]

import warnings

from erlab.io.dataloader import (
    load,
    loader_context,
    loaders,
    set_data_dir,
    set_loader,
    summarize,
)
from erlab.io.igor import load_experiment, load_wave
from erlab.io.utilities import load_hdf5, open_hdf5, save_as_hdf5, save_as_netcdf

# Import plugins last
# isort: off
import erlab.io.plugins  # noqa: F401

merlin = loaders["merlin"]
ssrl52 = loaders["ssrl52"]


def load_igor_ibw(*args, **kwargs):
    warnings.warn("Use `erlab.io.load_wave` instead", DeprecationWarning, stacklevel=2)
    return load_wave(*args, **kwargs)


def load_igor_pxp(*args, **kwargs):
    warnings.warn(
        "Use `erlab.io.load_experiment` instead", DeprecationWarning, stacklevel=2
    )
    return load_experiment(*args, **kwargs)



"""Provides core functionality of imagetool."""

from __future__ import annotations

__all__ = ["ImageSlicerArea"]

import collections
import functools
import gc
import inspect
import os
import time
import weakref
from typing import TYPE_CHECKING, Any

import numpy as np
import numpy.typing as npt
import pyqtgraph as pg
import xarray as xr
from qtpy import QtCore, QtGui, QtWidgets

from erlab.interactive.colors import (
    BetterColorBarItem,
    BetterImageItem,
    pg_colormap_powernorm,
)
from erlab.interactive.imagetool.slicer import ArraySlicer
from erlab.interactive.utilities import BetterAxisItem

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Sequence

    from pyqtgraph.graphicsItems.ViewBox import ViewBoxMenu
    from pyqtgraph.GraphicsScene import mouseEvents

suppressnanwarning = np.testing.suppress_warnings()
suppressnanwarning.filter(RuntimeWarning, r"All-NaN (slice|axis) encountered")


def _link_splitters(
    s0: QtWidgets.QSplitter, s1: QtWidgets.QSplitter, reverse: bool = False
):
    s0.blockSignals(True)
    s1.blockSignals(True)
    sizes = s0.sizes()
    total = sum(sizes)
    if reverse:
        sizes = list(reversed(sizes))
        sizes[0] = s1.sizes()[-1]
    else:
        sizes[0] = s1.sizes()[0]
    if all(x == 0 for x in sizes[1:]) and sizes[0] != total:
        sizes[1:] = [1] * len(sizes[1:])
    try:
        factor = (total - sizes[0]) / sum(sizes[1:])
    except ZeroDivisionError:
        factor = 0.0
    for k in range(1, len(sizes)):
        sizes[k] = round(sizes[k] * factor)
    if reverse:
        sizes = list(reversed(sizes))
    s0.setSizes(sizes)
    s0.blockSignals(False)
    s1.blockSignals(False)


def _sync_splitters(s0: QtWidgets.QSplitter, s1: QtWidgets.QSplitter):
    s0.splitterMoved.connect(lambda: _link_splitters(s1, s0))
    s1.splitterMoved.connect(lambda: _link_splitters(s0, s1))


# class ItoolGraphicsLayoutWidget(pg.GraphicsLayoutWidget):
class ItoolGraphicsLayoutWidget(pg.PlotWidget):
    # Unsure of whether to subclass GraphicsLayoutWidget or PlotWidget at the moment
    # Will need to run some benchmarks in the future
    def __init__(
        self,
        slicer_area: ImageSlicerArea,
        display_axis: tuple[int, ...],
        image: bool = False,
        **item_kw,
    ):
        # super().__init__()
        # self.ci.layout.setContentsMargins(0, 0, 0, 0)
        # self.ci.layout.setSpacing(3)
        # self.plotItem = ItoolPlotItem(slicer_area, display_axis, image, **item_kw)
        # self.addItem(self.plotItem)

        super().__init__(
            plotItem=ItoolPlotItem(slicer_area, display_axis, image, **item_kw),
        )
        self.viewport().setAttribute(
            QtCore.Qt.WidgetAttribute.WA_AcceptTouchEvents, False
        )
        self.scene().sigMouseClicked.connect(self.getPlotItem().mouseDragEvent)

    def getPlotItem(self) -> ItoolPlotItem:
        return self.plotItem

    def getPlotItemViewBox(self) -> pg.ViewBox:
        return self.getPlotItem().vb


def link_slicer(
    func: Callable | None = None,
    *,
    indices: bool = False,
    steps: bool = False,
    color: bool = False,
):
    """
    An internal decorator for choosing which functions to sync accross multiple
    instances of `ImageSlicerArea`.

    Parameters
    ----------
    func
        The method to sync across multiple instances of `ImageSlicerArea`.
    indices
        If `True`, the input argument named `value` given to `func` are interpreted as
        indices, and will be converted to appropriate values for other instances of
        `ImageSlicerArea`. The behavior of this conversion is determined by `steps`.
    steps
        If `False`, considers `value` as an absolute index. If `True`, considers
        `value` as a relative value such as the number of steps or bins. See the
        implementation of `SlicerLinkProxy` for more information.
    color
        Boolean whether the decorated method is related to visualization, such as
        colormap control.

    """

    def my_decorator(func: Callable):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            # skip sync if already synced
            skip_sync: bool = kwargs.pop("__slicer_skip_sync", False)

            out = func(*args, **kwargs)
            if args[0].is_linked:
                if not skip_sync:
                    all_args = inspect.Signature.from_callable(func).bind(
                        *args, **kwargs
                    )
                    all_args.apply_defaults()
                    obj: ImageSlicerArea = all_args.arguments.pop("self")
                    obj._linking_proxy.sync(
                        obj, func.__name__, all_args.arguments, indices, steps, color
                    )
            return out

        return wrapped

    if func is not None:
        return my_decorator(func)
    return my_decorator


class SlicerLinkProxy:
    """Internal class for handling linked `ImageSlicerArea` s.

    Parameters
    ----------
    link_colors
        Whether to sync color related changes.
    *slicers
        The slicers to link.

    """

    def __init__(self, *slicers: list[ImageSlicerArea], link_colors: bool = True):
        self.link_colors = link_colors
        self._slicers: set[ImageSlicerArea] = set()
        for s in slicers:
            self.add(s)

    def add(self, slicer: ImageSlicerArea):
        if slicer.is_linked:
            if slicer._linking_proxy == self:
                return
            else:
                raise ValueError("Already linked to another proxy.")
        self._slicers.add(slicer)
        slicer._linking_proxy = self

    def remove(self, slicer: ImageSlicerArea):
        self._slicers.remove(slicer)
        slicer._linking_proxy = None

    def sync(
        self,
        source: ImageSlicerArea,
        funcname: str,
        arguments: dict[str, Any],
        indices: bool,
        steps: bool,
        color: bool,
    ):
        """The core method that propagates changes across multiple `ImageSlicerArea`s.

        This method is invoked every time a method decorated with :func:`link_slicer` in
        a linked `ImageSlicerArea` is called.

        Parameters
        ----------
        source
            Instance of `ImageSlicerArea` corresponding to the called method.
        funcname
            Name of the called method.
        arguments
            Arguments included in the function call.
        indices, steps, color
            Arguments given to the decorator. See :func:`link_slicer`

        """
        if color and not self.link_colors:
            return
        for target in self._slicers.difference({source}):
            getattr(target, funcname)(
                **self.convert_args(source, target, arguments, indices, steps)
            )

    def convert_args(
        self,
        source: ImageSlicerArea,
        target: ImageSlicerArea,
        args: dict[str, Any],
        indices: bool,
        steps: bool,
    ):
        if indices:
            axis: int | None = args.get("axis")
            index: int | None = args.get("value")

            if index is not None:
                if axis is None:
                    args["value"] = [
                        self.convert_index(source, target, a, i, steps)
                        for (a, i) in zip(axis, index)
                    ]
                else:
                    args["value"] = self.convert_index(
                        source, target, axis, index, steps
                    )

        args["__slicer_skip_sync"] = True  # passed onto the decorator
        return args

    @staticmethod
    def convert_index(
        source: ImageSlicerArea,
        target: ImageSlicerArea,
        axis: int,
        index: int,
        steps: bool,
    ):
        if steps:
            return round(
                index * source.array_slicer.incs[axis] / target.array_slicer.incs[axis]
            )
        else:
            value: np.float32 = source.array_slicer.value_of_index(
                axis, index, uniform=False
            )
            new_index: int = target.array_slicer.index_of_value(
                axis, value, uniform=False
            )
            return new_index


class ImageSlicerArea(QtWidgets.QWidget):
    """A interactive tool based on :mod:`pyqtgraph` for exploring 3D data.

    Parameters
    ----------
    parent
        Parent widget.
    data
        Data to display. The data must have 2 to 4 dimensions.
    cmap
        Default colormap of the data.
    gamma
        Default power law normalization of the colormap.
    zeroCentered
        If `True`, the normalization is applied symmetrically from the midpoint of
        the colormap.
    rad2deg
        If `True` and `data` is not `None`, converts some known angle coordinates to
        degrees. If an iterable of strings is given, only the coordinates that
        correspond to the given strings are converted.
    bench
        Prints the fps on Ctrl + drag, for debug purposes

    Signals
    -------
    sigDataChanged()

    sigCurrentCursorChanged(index)

    sigViewOptionChanged()

    sigCursorCountChanged(n_cursors)
        Inherited from :class:`erlab.interactive.slicer.ArraySlicer`.
    sigIndexChanged(cursor, axes)
        Inherited from :class:`erlab.interactive.slicer.ArraySlicer`.
    sigBinChanged(cursor, axes)
        Inherited from :class:`erlab.interactive.slicer.ArraySlicer`.
    sigShapeChanged()
        Inherited from :class:`erlab.interactive.slicer.ArraySlicer`.

    """

    COLORS: list[QtGui.QColor] = [
        pg.mkColor(0.8),
        pg.mkColor("y"),
        pg.mkColor("m"),
        pg.mkColor("c"),
        pg.mkColor("g"),
        pg.mkColor("r"),
        pg.mkColor("b"),
    ]  #: List of :class:`PySide6.QtGui.QColor` containing colors for multiple cursors.

    sigDataChanged = QtCore.Signal()  #: :meta private:
    sigCurrentCursorChanged = QtCore.Signal(int)  #: :meta private:
    sigViewOptionChanged = QtCore.Signal()  #: :meta private:

    @property
    def sigCursorCountChanged(self) -> QtCore.SignalInstance:
        """:meta private:"""
        return self.array_slicer.sigCursorCountChanged

    @property
    def sigIndexChanged(self) -> QtCore.SignalInstance:
        """:meta private:"""
        return self.array_slicer.sigIndexChanged

    @property
    def sigBinChanged(self) -> QtCore.SignalInstance:
        """:meta private:"""
        return self.array_slicer.sigBinChanged

    @property
    def sigShapeChanged(self) -> QtCore.SignalInstance:
        """:meta private:"""
        return self.array_slicer.sigShapeChanged

    def __init__(
        self,
        parent: QtWidgets.QWidget | None = None,
        data: xr.DataArray | npt.ArrayLike | None = None,
        cmap: str | pg.ColorMap = "magma",
        gamma: float = 0.5,
        zeroCentered: bool = False,
        rad2deg: bool | Iterable[str] = False,
        *,
        bench: bool = False,
        image_cls=None,
        plotdata_cls=None,
    ):
        super().__init__(parent)

        self._linking_proxy: SlicerLinkProxy | None = None

        self.bench = bench

        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self._splitters = (
            QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical),
            QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal),
            QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical),
            QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical),
            QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal),
            QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical),
            QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal),
        )
        for s in self._splitters:
            s.setHandleWidth(4)
            s.setStyleSheet("QSplitter::handle{background: #222222;}")
            # palette = s.palette()
            # palette.setColor(QtGui.QPalette.ColorRole.Light, QtGui.QColor("yellow"))
            # s.setPalette(palette)
            # print(s.handleWidth())
            # pass
        self.layout().addWidget(self._splitters[0])
        for i, j in ((0, 1), (1, 2), (1, 3), (0, 4), (4, 5), (4, 6)):
            self._splitters[i].addWidget(self._splitters[j])
        _sync_splitters(self._splitters[1], self._splitters[4])

        self.cursor_colors: list[QtGui.QColor] = [self.COLORS[0]]

        self._colorbar = ItoolColorBar(self)
        self.layout().addWidget(self._colorbar)
        self._colorbar.setVisible(False)

        pkw = {"image_cls": image_cls, "plotdata_cls": plotdata_cls}
        self.manual_limits: dict[str | list[float]] = {}
        self._plots: tuple[ItoolGraphicsLayoutWidget, ...] = (
            ItoolGraphicsLayoutWidget(self, image=True, display_axis=(0, 1), **pkw),
            ItoolGraphicsLayoutWidget(self, display_axis=(0,), **pkw),
            ItoolGraphicsLayoutWidget(self, display_axis=(1,), is_vertical=True, **pkw),
            ItoolGraphicsLayoutWidget(self, display_axis=(2,), **pkw),
            ItoolGraphicsLayoutWidget(self, image=True, display_axis=(0, 2), **pkw),
            ItoolGraphicsLayoutWidget(self, image=True, display_axis=(2, 1), **pkw),
            ItoolGraphicsLayoutWidget(self, display_axis=(3,), **pkw),
            ItoolGraphicsLayoutWidget(self, image=True, display_axis=(3, 2), **pkw),
        )
        for i in (1, 4):
            self._splitters[2].addWidget(self._plots[i])
        for i in (6, 3, 7):
            self._splitters[3].addWidget(self._plots[i])
        self._splitters[5].addWidget(self._plots[0])
        for i in (5, 2):
            self._splitters[6].addWidget(self._plots[i])

        self.qapp: QtWidgets.QApplication = QtWidgets.QApplication.instance()
        self.qapp.aboutToQuit.connect(self.on_close)

        cmap_reversed = False

        if isinstance(cmap, str):
            if cmap.endswith("_r"):
                cmap = cmap[:-2]
                cmap_reversed = True
            if cmap.startswith("cet_CET"):
                cmap = cmap[4:]

        self.colormap_properties: dict[str, str | pg.ColorMap | float | bool] = {
            "cmap": cmap,
            "gamma": gamma,
            "reversed": cmap_reversed,
            "highContrast": False,
            "zeroCentered": zeroCentered,
        }

        self._data: xr.DataArray | None = None
        self.current_cursor: int = 0

        self.levels_locked: bool = False

        if data is not None:
            self.set_data(data, rad2deg=rad2deg)

        if self.bench:
            print("\n")

    def on_close(self):
        self.array_slicer.clear_cache()
        self.data.close()
        if hasattr(self, "_data"):
            self._data.close()
            del self._data
        gc.collect()

    def connect_axes_signals(self):
        for ax in self.axes:
            ax.connect_signals()

    def disconnect_axes_signals(self):
        for ax in self.axes:
            ax.disconnect_signals()

    def connect_signals(self):
        self.connect_axes_signals()
        self.sigDataChanged.connect(self.refresh_all)
        self.sigShapeChanged.connect(self.refresh_all)
        self.sigCursorCountChanged.connect(lambda: self.set_colormap(update=True))

    def add_link(self, proxy: SlicerLinkProxy):
        proxy.add(self)

    def remove_link(self):
        self._linking_proxy.remove(self)

    @property
    def is_linked(self) -> bool:
        return self._linking_proxy is not None

    @property
    def colormap(self) -> str | pg.ColorMap:
        return self.colormap_properties["cmap"]

    @property
    def slices(self) -> tuple[ItoolPlotItem, ...]:
        if self.data.ndim == 2:
            return ()
        elif self.data.ndim == 3:
            return tuple(self.get_axes(ax) for ax in (4, 5))
        elif self.data.ndim == 4:
            return tuple(self.get_axes(ax) for ax in (4, 5, 7))

    @property
    def profiles(self) -> tuple[ItoolPlotItem, ...]:
        if self.data.ndim == 2:
            profile_axes = (1, 2)
        elif self.data.ndim == 3:
            profile_axes = (1, 2, 3)
        else:
            profile_axes = (1, 2, 3, 6)
        return tuple(self.get_axes(ax) for ax in profile_axes)

    @property
    def main_image(self) -> ItoolPlotItem:
        """returns the main PlotItem"""
        return self.get_axes(0)

    @property
    def images(self) -> tuple[ItoolPlotItem, ...]:
        return (self.main_image, *self.slices)

    @property
    def axes(self) -> tuple[ItoolPlotItem, ...]:
        """Currently valid subset of self._plots"""
        return self.images + self.profiles

    @property
    def _imageitems(self) -> tuple[ItoolImageItem, ...]:
        return tuple(im for ax in self.images for im in ax.slicer_data_items)

    @property
    def array_slicer(self) -> ArraySlicer:
        return self._array_slicer

    @property
    def n_cursors(self) -> int:
        return self.array_slicer.n_cursors

    @property
    def current_indices(self) -> list[int]:
        return self.array_slicer.get_indices(self.current_cursor)

    @property
    def current_values(self) -> list[float]:
        return self.array_slicer.get_values(self.current_cursor, uniform=False)

    @property
    def current_values_uniform(self) -> list[float]:
        return self.array_slicer.get_values(self.current_cursor, uniform=True)

    @property
    def data(self) -> xr.DataArray:
        return self.array_slicer._obj

    @property
    def color_locked(self) -> bool:
        return self._colorbar.isVisible()

    def get_current_index(self, axis: int) -> int:
        return self.array_slicer.get_index(self.current_cursor, axis)

    def get_current_value(self, axis: int, uniform: bool = False) -> float:
        return self.array_slicer.get_value(self.current_cursor, axis, uniform=uniform)

    def get_axes_widget(self, index: int) -> ItoolGraphicsLayoutWidget:
        return self._plots[index]

    def get_axes(self, index: int) -> ItoolPlotItem:
        return self._plots[index].plotItem

    @QtCore.Slot()
    @QtCore.Slot(tuple)
    @QtCore.Slot(tuple, bool)
    def refresh_all(
        self, axes: tuple[int, ...] | None = None, only_plots: bool = False
    ):
        for c in range(self.n_cursors):
            self.sigIndexChanged.emit(c, axes)
        if not only_plots:
            for ax in self.axes:
                ax.refresh_labels()

    @QtCore.Slot(tuple)
    @link_slicer
    def refresh_current(self, axes: tuple[int, ...] | None = None):
        self.sigIndexChanged.emit(self.current_cursor, axes)

    @QtCore.Slot(int, list)
    @link_slicer
    def refresh(self, cursor: int, axes: tuple[int, ...] | None = None):
        self.sigIndexChanged.emit(cursor, axes)

    def view_all(self):
        for ax in self.axes:
            ax.vb.enableAutoRange()
            ax.vb.updateAutoRange()

    @link_slicer
    def center_all_cursors(self):
        for i in range(self.n_cursors):
            self.array_slicer.center_cursor(i)

    @link_slicer
    def center_cursor(self):
        self.array_slicer.center_cursor(self.current_cursor)

    @link_slicer
    def set_current_cursor(self, cursor: int, update=True):
        if cursor > self.n_cursors - 1:
            raise IndexError("Cursor index out of range")
        self.current_cursor = cursor
        if update:
            self.refresh_current()
        self.sigCurrentCursorChanged.emit(cursor)

    def set_data(
        self, data: xr.DataArray | npt.ArrayLike, rad2deg: bool | Iterable[str] = False
    ):
        if hasattr(self, "_array_slicer"):
            n_cursors_old = self.n_cursors
            if isinstance(self._data, xr.DataArray):
                self._data.close()
            del self._data
            self.disconnect_axes_signals()
        else:
            n_cursors_old = 1

        if not isinstance(data, xr.DataArray):
            if isinstance(data, xr.Dataset):
                try:
                    data = data.spectrum
                except AttributeError:
                    data = data[next(iter(data.data_vars.keys()))]
            else:
                data = xr.DataArray(np.asarray(data))

        if not rad2deg:
            self._data = data
        else:
            if np.iterable(rad2deg):
                conv_dims = rad2deg
            else:
                conv_dims = [
                    d
                    for d in ("phi", "theta", "beta", "alpha", "chi", "xi")
                    if d in data.dims
                ]
            self._data = data.assign_coords({d: np.rad2deg(data[d]) for d in conv_dims})

        if hasattr(self, "_array_slicer"):
            self._array_slicer.set_array(self._data, reset=True)
        else:
            self._array_slicer = ArraySlicer(self._data)

        while self.n_cursors != n_cursors_old:
            self.array_slicer.add_cursor(update=False)

        self.connect_signals()

        self.adjust_layout()

        if self.current_cursor > self.n_cursors - 1:
            self.set_current_cursor(self.n_cursors - 1, update=False)
        self.sigDataChanged.emit()

        # self.refresh_current()
        self.set_colormap(update=True)
        self._colorbar.cb.setImageItem()
        self.lock_levels(False)

    @QtCore.Slot(int, int)
    @link_slicer
    def swap_axes(self, ax1: int, ax2: int):
        self.array_slicer.swap_axes(ax1, ax2)

    @QtCore.Slot(int, int, bool)
    @link_slicer(indices=True)
    def set_index(
        self, axis: int, value: int, update: bool = True, cursor: int | None = None
    ):
        if cursor is None:
            cursor = self.current_cursor
        self.array_slicer.set_index(cursor, axis, value, update)

    @QtCore.Slot(int, int, bool)
    @link_slicer(indices=True, steps=True)
    def step_index(
        self, axis: int, value: int, update: bool = True, cursor: int | None = None
    ):
        if cursor is None:
            cursor = self.current_cursor
        self.array_slicer.step_index(cursor, axis, value, update)

    @QtCore.Slot(int, int, bool)
    @link_slicer(indices=True, steps=True)
    def step_index_all(self, axis: int, value: int, update: bool = True):
        for i in range(self.n_cursors):
            self.array_slicer.step_index(i, axis, value, update)

    @QtCore.Slot(int, float, bool, bool)
    @link_slicer
    def set_value(
        self,
        axis: int,
        value: float,
        update: bool = True,
        uniform: bool = False,
        cursor: int | None = None,
    ):
        if cursor is None:
            cursor = self.current_cursor
        self.array_slicer.set_value(cursor, axis, value, update, uniform)

    @QtCore.Slot(int, int, bool)
    @link_slicer(indices=True, steps=True)
    def set_bin(
        self, axis: int, value: int, update: bool = True, cursor: int | None = None
    ):
        if cursor is None:
            cursor = self.current_cursor
        new_bins: list[int | None] = [None] * self.data.ndim
        new_bins[axis] = value
        self.array_slicer.set_bins(cursor, new_bins, update)

    @QtCore.Slot(int, int, bool)
    @link_slicer(indices=True, steps=True)
    def set_bin_all(self, axis: int, value: int, update: bool = True):
        new_bins: list[int | None] = [None] * self.data.ndim
        new_bins[axis] = value
        for c in range(self.n_cursors):
            self.array_slicer.set_bins(c, new_bins, update)

    @QtCore.Slot()
    @link_slicer
    def add_cursor(self):
        self.array_slicer.add_cursor(self.current_cursor, update=False)
        self.cursor_colors.append(self.gen_cursor_color(self.n_cursors - 1))
        self.current_cursor = self.n_cursors - 1
        for ax in self.axes:
            ax.add_cursor(update=False)
        self._colorbar.cb.level_change()
        self.refresh_current()
        self.sigCursorCountChanged.emit(self.n_cursors)
        self.sigCurrentCursorChanged.emit(self.current_cursor)

    @QtCore.Slot(int)
    @link_slicer
    def remove_cursor(self, index: int):
        if self.n_cursors == 1:
            return
        self.array_slicer.remove_cursor(index, update=False)
        self.cursor_colors.pop(index)
        if self.current_cursor == index:
            if index == 0:
                self.current_cursor = 1
            self.current_cursor -= 1
        for ax in self.axes:
            ax.remove_cursor(index)
        self.refresh_current()
        self.sigCursorCountChanged.emit(self.n_cursors)
        self.sigCurrentCursorChanged.emit(self.current_cursor)

    @QtCore.Slot()
    def remove_current_cursor(self):
        self.remove_cursor(self.current_cursor)

    def gen_cursor_color(self, index: int) -> QtGui.QColor:
        clr = self.COLORS[index % len(self.COLORS)]
        while clr in self.cursor_colors:
            clr = self.COLORS[index % len(self.COLORS)]
            if self.n_cursors > len(self.COLORS):
                break
            index += 1
        return clr

    def gen_cursor_colors(
        self, index: int
    ) -> tuple[QtGui.QColor, QtGui.QColor, QtGui.QColor, QtGui.QColor, QtGui.QColor]:
        clr = self.cursor_colors[index]

        clr_cursor = pg.mkColor(clr)
        clr_cursor_hover = pg.mkColor(clr)
        clr_span = pg.mkColor(clr)
        clr_span_edge = pg.mkColor(clr)

        clr_cursor.setAlphaF(0.75)
        clr_cursor_hover.setAlphaF(0.95)
        clr_span.setAlphaF(0.15)
        clr_span_edge.setAlphaF(0.35)

        return clr, clr_cursor, clr_cursor_hover, clr_span, clr_span_edge

    @link_slicer(color=True)
    def set_colormap(
        self,
        cmap: str | pg.ColorMap | None = None,
        gamma: float | None = None,
        reversed: bool | None = None,
        highContrast: bool | None = None,
        zeroCentered: bool | None = None,
        update: bool = True,
    ):
        if cmap is not None:
            self.colormap_properties["cmap"] = cmap
        if gamma is not None:
            self.colormap_properties["gamma"] = gamma
        if reversed is not None:
            self.colormap_properties["reversed"] = reversed
        if highContrast is not None:
            self.colormap_properties["highContrast"] = highContrast
        if zeroCentered is not None:
            self.colormap_properties["zeroCentered"] = zeroCentered

        cmap = pg_colormap_powernorm(
            self.colormap_properties["cmap"],
            self.colormap_properties["gamma"],
            self.colormap_properties["reversed"],
            highContrast=self.colormap_properties["highContrast"],
            zeroCentered=self.colormap_properties["zeroCentered"],
        )
        for im in self._imageitems:
            im.set_pg_colormap(cmap, update=update)
        self.sigViewOptionChanged.emit()

    @QtCore.Slot(bool)
    def lock_levels(self, lock: bool):
        self.levels_locked: bool = lock

        if self.levels_locked:
            levels = self.array_slicer.limits
            self._colorbar.cb.setLimits(levels)
        for im in self._imageitems:
            if self.levels_locked:
                im.setLevels(levels, update=False)
            else:
                im.levels = None
            im.refresh_data()

        self._colorbar.setVisible(self.levels_locked)
        self.sigViewOptionChanged.emit()

    def adjust_layout(
        self,
        horiz_pad: int = 45,
        vert_pad: int = 30,
        font_size: float = 11.0,
        r: tuple[float, float, float, float] = (1.2, 1.5, 3.0, 1.0),
    ):
        """Determines the padding and aspect ratios.

        Parameters
        ----------
        horiz_pad, vert_pad
            Reserved space for the x and y axes.
        font_size
            Font size in points.
        r
            4 numbers that determine the layout aspect ratios. See notes.

        Notes
        -----
        Axes indices and layout parameters.

        .. code-block:: text

                 ┌───────────┬───────────┐
            r[0] │     1     │     6     │
                 │           ├───────────┤
                 ├───────────┤     3     │
            r[1] │     4     ├───────────┤
                 │           │     7     │
                 │───────────┼───────┬───┤
                 │           │       │   │
            r[2] │     0     │   5   │ 2 │
                 │           │       │   │
                 └───────────┴───────┴───┘
                  r[3] * r[2]

        """

        font = QtGui.QFont()
        font.setPointSizeF(float(font_size))

        valid_axis: tuple[tuple[bool, bool, bool, bool]] = (
            (1, 0, 0, 1),
            (1, 1, 0, 0),
            (0, 0, 1, 1),
            (0, 1, 1, 0),
            (1, 0, 0, 0),
            (0, 0, 0, 1),
            (0, 1, 1, 0),
            (0, 1, 1, 0),
        )  # booleans corresponding to the (left, top, right, bottom) axes of each plot.

        invalid: list[int] = []  # axes to hide.
        r0, r1, r2, r3 = r

        # !TODO: automate this based on ItoolPlotItem.display_axis
        if self.data.ndim == 2:
            invalid = [4, 5, 6, 7]
            r1 = r0 / 6
        elif self.data.ndim == 3:
            invalid = [6, 7]

        r01 = r0 / r1
        scale = 100
        d = self._splitters[0].handleWidth() / scale  # padding due to splitters
        sizes: list[tuple[float, ...]] = [
            (r0 + r1, r2),
            (r3 * r2, r3 * (r0 + r1)),
            ((r0 + r1 - d) * r01, (r0 + r1 - d) / r01),
            ((r0 + r1 - d) / 2, (r0 + r1 - d) / 2, 0),
            (r3 * r2, r3 * (r0 + r1)),
            (r2,),
            ((r3 * (r0 + r1) - d) / r01, (r3 * (r0 + r1) - d) * r01),
        ]
        if self.data.ndim == 4:
            sizes[3] = (0, 0, (r0 + r1 - d))
        for split, sz in zip(self._splitters, sizes):
            split.setSizes(tuple(round(s * scale) for s in sz))

        for i, sel in enumerate(valid_axis):
            self.get_axes_widget(i).setVisible(i not in invalid)
            if i in invalid:
                continue
            axes = self.get_axes(i)
            axes.setDefaultPadding(0)
            for axis in ("left", "bottom", "right", "top"):
                axes.getAxis(axis).setTickFont(font)
                axes.getAxis(axis).setStyle(
                    autoExpandTextSpace=True, autoReduceTextSpace=True
                )
            axes.showAxes(sel, showValues=sel, size=(horiz_pad, vert_pad))
            if i in (1, 4):
                axes.setXLink(self.get_axes(0))
            elif i in (2, 5):
                axes.setYLink(self.get_axes(0))

        # reserve space, only hide plotItem
        self.get_axes(3).setVisible(self.data.ndim != 2)

        self._colorbar.set_dimensions(
            width=horiz_pad + 30, horiz_pad=None, vert_pad=vert_pad, font_size=font_size
        )

    def toggle_snap(self, value: bool | None = None):
        if value is None:
            value = not self.array_slicer.snap_to_data
        elif value == self.array_slicer.snap_to_data:
            return
        self.array_slicer.snap_to_data = value
        self.sigViewOptionChanged.emit()

    def changeEvent(self, evt: QtCore.QEvent):
        if evt.type() == QtCore.QEvent.Type.PaletteChange:
            self.qapp.setStyle(self.qapp.style().name())
        super().changeEvent(evt)


class ItoolCursorLine(pg.InfiniteLine):
    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)
        self.qapp: QtWidgets.QApplication = QtWidgets.QApplication.instance()

    @property
    def plotItem(self) -> ItoolPlotItem:
        return self.parentItem().parentItem().parentItem()

    def setBounds(self, bounds: Sequence[float], value: float | None = None):
        if bounds[0] > bounds[1]:
            bounds = list(bounds)
            bounds.reverse()
        self.maxRange = bounds
        if value is None:
            value = self.value()
        self.setValue(value)

    def value(self) -> float:
        return float(super().value())

    def mouseDragEvent(self, ev: mouseEvents.MouseDragEvent):
        if (
            QtCore.Qt.KeyboardModifier.ControlModifier
            not in self.qapp.keyboardModifiers()
        ):
            if self.movable and ev.button() == QtCore.Qt.MouseButton.LeftButton:
                if ev.isStart():
                    self.moving = True
                    self.cursorOffset = self.pos() - self.mapToParent(
                        ev.buttonDownPos()
                    )
                    self.startPosition = self.pos()
                ev.accept()

                if not self.moving:
                    return

                new_position = self.cursorOffset + self.mapToParent(ev.pos())
                if self.angle % 180 == 0:
                    self.temp_value = new_position.y()
                elif self.angle % 180 == 90:
                    self.temp_value = new_position.x()

                self.sigDragged.emit(self)
                if ev.isFinish():
                    self.moving = False
                    self.sigPositionChangeFinished.emit(self)
        else:
            self.setMouseHover(False)
            self.plotItem.mouseDragEvent(ev)

    def mouseClickEvent(self, ev: mouseEvents.MouseClickEvent):
        if (
            QtCore.Qt.KeyboardModifier.ControlModifier
            not in self.qapp.keyboardModifiers()
        ):
            super().mouseClickEvent(ev)
        else:
            self.setMouseHover(False)
            ev.ignore()

    def hoverEvent(self, ev):
        if (
            QtCore.Qt.KeyboardModifier.ControlModifier
            not in self.qapp.keyboardModifiers()
        ):
            super().hoverEvent(ev)
        else:
            self.setMouseHover(False)


class ItoolCursorSpan(pg.LinearRegionItem):
    def __init__(self, *args, **kargs):
        kargs.setdefault("movable", False)
        super().__init__(*args, **kargs)

    def setRegion(self, rgn):
        # hides when region width is 0
        if rgn[1] == rgn[0]:
            self.setVisible(False)
        else:
            self.setVisible(True)
            super().setRegion(rgn)


class ItoolDisplayObject:
    def __init__(self, axes, cursor: int | None = None):
        self.axes = axes
        if cursor is None:
            cursor = 0
        self._cursor_index = int(cursor)
        self.qapp: QtGui.QGuiApplication = QtGui.QGuiApplication.instance()

    @property
    def display_axis(self):
        return self.axes.display_axis

    @property
    def slicer_area(self) -> ImageSlicerArea:
        return self.axes.slicer_area

    @property
    def array_slicer(self) -> ArraySlicer:
        return self.axes.array_slicer

    @property
    def cursor_index(self):
        return self._cursor_index

    @cursor_index.setter
    def cursor_index(self, value: int):
        self._cursor_index = int(value)

    @property
    def sliced_data(self) -> xr.DataArray:
        return self.array_slicer.xslice(self.cursor_index, self.display_axis)

    def refresh_data(self):
        pass


class ItoolPlotDataItem(ItoolDisplayObject, pg.PlotDataItem):
    def __init__(
        self,
        axes,
        cursor: int | None = None,
        is_vertical: bool = False,
        **kargs,
    ):
        pg.PlotDataItem.__init__(self, axes=axes, cursor=cursor, **kargs)
        ItoolDisplayObject.__init__(self, axes=axes, cursor=cursor)
        self.is_vertical = is_vertical

    def refresh_data(self):
        ItoolDisplayObject.refresh_data(self)
        coord, vals = self.array_slicer.slice_with_coord(
            self.cursor_index, self.display_axis
        )
        if self.is_vertical:
            self.setData(vals, coord)
        else:
            self.setData(coord, vals)


class ItoolImageItem(ItoolDisplayObject, BetterImageItem):
    def __init__(
        self,
        axes,
        cursor: int | None = None,
        **kargs,
    ):
        BetterImageItem.__init__(self, axes=axes, cursor=cursor, **kargs)
        ItoolDisplayObject.__init__(self, axes=axes, cursor=cursor)

    def updateImage(self, *args, **kargs):
        defaults = {"autoLevels": not self.slicer_area.levels_locked}
        defaults.update(kargs)
        return self.setImage(*args, **defaults)

    @suppressnanwarning
    def refresh_data(self):
        ItoolDisplayObject.refresh_data(self)
        rect, img = self.array_slicer.slice_with_coord(
            self.cursor_index, self.display_axis
        )
        self.setImage(
            image=img, rect=rect, autoLevels=not self.slicer_area.levels_locked
        )

    def mouseDragEvent(self, ev: mouseEvents.MouseDragEvent):
        if QtCore.Qt.KeyboardModifier.ControlModifier in self.qapp.keyboardModifiers():
            ev.ignore()
        else:
            super().mouseDragEvent(ev)

    def mouseClickEvent(self, ev: mouseEvents.MouseClickEvent):
        if QtCore.Qt.KeyboardModifier.ControlModifier in self.qapp.keyboardModifiers():
            ev.ignore()
        else:
            super().mouseClickEvent(ev)


class ItoolPlotItem(pg.PlotItem):
    sigDragged = QtCore.Signal(object, object)

    def __init__(
        self,
        slicer_area: ImageSlicerArea,
        display_axis: tuple[int, ...],
        image: bool = False,
        image_cls=None,
        plotdata_cls=None,
        **item_kw,
    ):
        super().__init__(
            axisItems={a: BetterAxisItem(a) for a in ("left", "right", "top", "bottom")}
        )
        for act in ["Transforms", "Downsample", "Average", "Alpha", "Points"]:
            self.setContextMenuActionVisible(act, False)

        save_action = self.vb.menu.addAction("Save data as HDF5")
        save_action.triggered.connect(lambda: self.save_current_data())

        for i in (0, 1):
            self.getViewBoxMenu().ctrl[i].linkCombo.setVisible(False)
            self.getViewBoxMenu().ctrl[i].label.setVisible(False)
        self.getViewBox().setCursor(QtGui.QCursor(QtCore.Qt.CursorShape.CrossCursor))

        self.slicer_area = slicer_area
        self._display_axis = display_axis

        self.is_image = image
        self._item_kw = item_kw

        if image_cls is None:
            self.image_cls = ItoolImageItem
        if plotdata_cls is None:
            self.plotdata_cls = ItoolPlotDataItem

        self.slicer_data_items: list[ItoolImageItem | ItoolPlotDataItem] = []
        self.cursor_lines: list[dict[int, ItoolCursorLine]] = []
        self.cursor_spans: list[dict[int, ItoolCursorSpan]] = []
        self.add_cursor(update=False)

        self.proxy = pg.SignalProxy(
            self.sigDragged,
            delay=1 / 60,
            rateLimit=60,
            slot=self.process_drag,
        )
        if self.slicer_area.bench:
            self._time_start = None
            self._time_end = None
            self._single_queue = collections.deque([0], maxlen=9)
            self._next_queue = collections.deque([0], maxlen=9)

    @property
    def axis_dims(self) -> list[str]:
        dim_list = [self.slicer_area.data.dims[ax] for ax in self.display_axis]
        if not self.is_image:
            if self.slicer_data_items[-1].is_vertical:
                dim_list = [None, *dim_list]
            else:
                dim_list = [*dim_list, None]
        return dim_list

    @property
    def is_independent(self) -> bool:
        return self.vb.state["linkedViews"] == [None, None]

    def refresh_manual_range(self):
        if self.is_independent:
            return
        for dim, auto, rng in zip(
            self.axis_dims, self.vb.state["autoRange"], self.vb.state["viewRange"]
        ):
            if dim is not None:
                if auto:
                    self.slicer_area.manual_limits.pop("dim", None)
                else:
                    self.slicer_area.manual_limits[dim] = rng

    def update_manual_range(self):
        if self.is_independent:
            return
        self.set_range_from(self.slicer_area.manual_limits)

    def set_range_from(self, limits: dict[str, list[float]], **kwargs: dict):
        for dim, key in zip(self.axis_dims, ("xRange", "yRange")):
            if dim is not None:
                try:
                    kwargs[key] = limits[dim]
                except KeyError:
                    pass
        if len(kwargs) != 0:
            self.setRange(**kwargs)

    def getMenu(self) -> QtWidgets.QMenu:
        return self.ctrlMenu

    def getViewBox(self) -> pg.ViewBox:
        return self.vb

    def getViewBoxMenu(self) -> ViewBoxMenu:
        return self.getViewBox().menu

    def mouseDragEvent(self, ev: mouseEvents.MouseDragEvent):
        modifiers = self.slicer_area.qapp.keyboardModifiers()
        if (
            QtCore.Qt.KeyboardModifier.ControlModifier in modifiers
            and ev.button() == QtCore.Qt.MouseButton.LeftButton
        ):
            ev.accept()
            self.sigDragged.emit(ev, modifiers)
            # self.process_drag((ev, modifiers))
        else:
            ev.ignore()

    def process_drag(
        self, sig: tuple[mouseEvents.MouseDragEvent, QtCore.Qt.KeyboardModifier]
    ):
        if self.slicer_area.bench:
            if self._time_end is not None:
                self._single_queue.append(1 / (self._time_end - self._time_start))
            self._time_end = self._time_start
            self._time_start = time.perf_counter()
            if self._time_end is not None:
                self._next_queue.append(1 / (self._time_start - self._time_end))
            print(
                "\x1b[2K\x1b[1A\x1b[2K"
                f"SingleUpdate\t{np.mean(self._single_queue):.1f}"
                f" ± {np.std(self._single_queue):.1f}",
                f"NextUpdate\t{np.mean(self._next_queue):.1f}"
                f" ± {np.std(self._next_queue):.1f}",
                sep="\t",
            )
            self._time_start = time.perf_counter()

        ev, modifiers = sig
        data_pos = self.getViewBox().mapSceneToView(ev.scenePos())

        if (not self.is_image) and self.slicer_data_items[-1].is_vertical:
            data_pos_coords = (data_pos.y(), data_pos.x())
        else:
            data_pos_coords = (data_pos.x(), data_pos.y())

        if QtCore.Qt.KeyboardModifier.AltModifier in modifiers:
            for c in range(self.slicer_area.n_cursors):
                for i, ax in enumerate(self.display_axis):
                    self.slicer_area.set_value(
                        ax, data_pos_coords[i], update=False, uniform=True, cursor=c
                    )
                self.slicer_area.refresh(c, self.display_axis)
        else:
            for i, ax in enumerate(self.display_axis):
                self.slicer_area.set_value(
                    ax, data_pos_coords[i], update=False, uniform=True
                )
            self.slicer_area.refresh_current(self.display_axis)

        if self.slicer_area.bench:
            self._time_end = time.perf_counter()

    def add_cursor(self, update=True):
        new_cursor = len(self.slicer_data_items)
        line_angles = (90, 0)

        (
            clr,
            clr_cursor,
            clr_cursor_hover,
            clr_span,
            clr_span_edge,
        ) = self.slicer_area.gen_cursor_colors(new_cursor)

        if self.is_image:
            item = self.image_cls(
                self,
                cursor=new_cursor,
                autoDownsample=True,
                axisOrder="row-major",
                **self._item_kw,
            )
            if self.slicer_area.color_locked:
                item.setLevels(self.array_slicer.limits, update=True)
        else:
            item = self.plotdata_cls(
                self,
                cursor=new_cursor,
                pen=pg.mkPen(pg.mkColor(clr)),
                defaultPadding=0.0,
                clipToView=False,
                connect="auto",
                **self._item_kw,
            )
            if item.is_vertical:
                line_angles = (0, 90)
        self.slicer_data_items.append(item)
        self.addItem(item)

        cursors: list[ItoolCursorLine] = []
        spans: list[ItoolCursorSpan] = []
        for ang in line_angles:
            cursors.append(
                ItoolCursorLine(
                    pen=pg.mkPen(clr_cursor, width=1),
                    hoverPen=pg.mkPen(clr_cursor_hover, width=1),
                    angle=ang,
                    movable=True,
                )
            )
            spans.append(
                ItoolCursorSpan(
                    orientation="vertical" if ang == 90 else "horizontal",
                    movable=False,
                    pen=pg.mkPen(clr_span_edge),
                    brush=pg.mkBrush(clr_span),
                )
            )

        self.cursor_lines.append({})
        self.cursor_spans.append({})
        for c, s, ax in zip(cursors, spans, self.display_axis):
            self.cursor_lines[-1][ax] = c
            self.cursor_spans[-1][ax] = s
            self.addItem(c)
            c.setZValue(10)
            self.addItem(s)
            s.setZValue(9)
            c.sigDragged.connect(
                lambda v, *, line=c, axis=ax: self.line_drag(line, v.temp_value, axis)
            )
            c.sigClicked.connect(lambda *, line=c: self.line_click(line))

        if update:
            self.refresh_cursor(new_cursor)

    def index_of_line(self, line: ItoolCursorLine) -> int:
        for i, line_dict in enumerate(self.cursor_lines):
            for v in line_dict.values():
                if v == line:
                    return i
        raise ValueError("`line` is not a valid cursor.")

    def line_click(self, line: ItoolCursorLine):
        cursor = self.index_of_line(line)
        if cursor != self.slicer_area.current_cursor:
            self.slicer_area.set_current_cursor(cursor, update=True)

    def line_drag(self, line: ItoolCursorLine, value: float, axis: int):
        cursor = self.index_of_line(line)
        if cursor != self.slicer_area.current_cursor:
            self.slicer_area.set_current_cursor(cursor, update=True)
        if (
            self.slicer_area.qapp.keyboardModifiers()
            != QtCore.Qt.KeyboardModifier.AltModifier
        ):
            self.slicer_area.set_value(
                axis, value, update=True, uniform=True, cursor=cursor
            )
        else:
            for i in range(self.slicer_area.n_cursors):
                self.slicer_area.set_value(
                    axis, value, update=True, uniform=True, cursor=i
                )

    def remove_cursor(self, index: int):
        item = self.slicer_data_items.pop(index)
        self.removeItem(item)
        for line, span in zip(
            self.cursor_lines.pop(index).values(), self.cursor_spans.pop(index).values()
        ):
            self.removeItem(line)
            self.removeItem(span)
        for i, item in enumerate(self.slicer_data_items):
            item.cursor_index = i

    def refresh_cursor(self, cursor: int):
        for ax, line in self.cursor_lines[cursor].items():
            line.setBounds(
                self.array_slicer.lims_uniform[ax],
                self.array_slicer.get_value(cursor, ax, uniform=True),
            )
            self.cursor_spans[cursor][ax].setRegion(
                self.array_slicer.span_bounds(cursor, ax)
            )

    def connect_signals(self):
        self._slicer_area.sigIndexChanged.connect(self.refresh_items_data)
        self._slicer_area.sigBinChanged.connect(self.refresh_items_data)
        self._slicer_area.sigShapeChanged.connect(self.update_manual_range)
        self.vb.sigRangeChanged.connect(self.refresh_manual_range)

    def disconnect_signals(self):
        self._slicer_area.sigIndexChanged.disconnect(self.refresh_items_data)
        self._slicer_area.sigBinChanged.disconnect(self.refresh_items_data)
        self._slicer_area.sigShapeChanged.disconnect(self.update_manual_range)
        self.vb.sigRangeChanged.disconnect(self.refresh_manual_range)

    @QtCore.Slot(int, object)
    def refresh_items_data(self, cursor: int, axes: tuple[int] | None = None):
        self.refresh_cursor(cursor)
        if axes is not None:
            # display_axis는 축 dim 표시하는거임. 즉 해당 축만 바뀌면 데이터 변화 없음
            if all(elem in self.display_axis for elem in axes):
                return
        for item in self.slicer_data_items:
            if item.cursor_index != cursor:
                continue
            self.set_active_cursor(cursor)
            item.refresh_data()
        # TODO: autorange smarter
        self.vb.updateAutoRange()

    @QtCore.Slot()
    def refresh_labels(self):
        if self.is_image:
            label_kw = {
                a: self._get_label_unit(i)
                for a, i in zip(("top", "bottom", "left", "right"), (0, 0, 1, 1))
                if self.getAxis(a).isVisible()
            }
        else:
            label_kw = {}

            if self.slicer_data_items[-1].is_vertical:
                valid_ax = ("left", "right")
            else:
                valid_ax = ("top", "bottom")

            for a in ("top", "bottom", "left", "right"):
                if self.getAxis(a).isVisible():
                    label_kw[a] = self._get_label_unit(0 if a in valid_ax else None)
        self.setLabels(**label_kw)

    def _get_label_unit(self, index: int | None = None):
        if index is None:
            return "", ""
        dim = self.slicer_area.data.dims[self.display_axis[index]]
        if dim == "eV":
            return dim, "eV"
        # if dim in ("kx", "ky", "phi", "theta", "beta", "alpha", "chi"):
        # return dim
        return dim, ""

    def set_active_cursor(self, index: int):
        if self.is_image:
            for i, item in enumerate(self.slicer_data_items):
                item.setVisible(i == index)

    def save_current_data(self, fileName=None):
        if fileName is None:
            self.fileDialog = QtWidgets.QFileDialog()
            self.fileDialog.setFileMode(QtWidgets.QFileDialog.FileMode.AnyFile)
            self.fileDialog.setAcceptMode(QtWidgets.QFileDialog.AcceptMode.AcceptSave)
            self.fileDialog.setNameFilter("xarray HDF5 Files (*.h5)")
            if pg.PlotItem.lastFileDir is not None:
                self.fileDialog.setDirectory(
                    os.path.join(
                        pg.PlotItem.lastFileDir, f"{self.slicer_area._data.name}.h5"
                    )
                )
            else:
                self.fileDialog.setDirectory(
                    os.path.join(os.getcwd(), f"{self.slicer_area._data.name}.h5")
                )
            self.fileDialog.show()
            self.fileDialog.fileSelected.connect(self.save_current_data)
            return

        fileName = str(fileName)
        pg.PlotItem.lastFileDir = os.path.dirname(fileName)

        import erlab.io

        erlab.io.save_as_hdf5(
            self.slicer_data_items[self.slicer_area.current_cursor].sliced_data,
            fileName,
        )

    @property
    def display_axis(self) -> tuple[int, ...]:
        return self._display_axis

    @display_axis.setter
    def display_axis(self, value: tuple[int, ...]):
        self._display_axis = value

    @property
    def slicer_area(self) -> ImageSlicerArea:
        return self._slicer_area

    @slicer_area.setter
    def slicer_area(self, value: ImageSlicerArea):
        self._slicer_area = value

    @property
    def array_slicer(self) -> ArraySlicer:
        return self.slicer_area.array_slicer


class ItoolColorBarItem(BetterColorBarItem):
    def __init__(self, slicer_area: ImageSlicerArea | None = None, **kwargs):
        self._slicer_area = slicer_area
        kwargs.setdefault(
            "axisItems",
            {a: BetterAxisItem(a) for a in ("left", "right", "top", "bottom")},
        )
        super().__init__(**kwargs)

    @property
    def slicer_area(self) -> ImageSlicerArea:
        return self._slicer_area

    @property
    def images(self):
        return [weakref.ref(x) for x in self._slicer_area._imageitems]

    @property
    def primary_image(self):
        return weakref.ref(self._slicer_area.main_image.slicer_data_items[0])

    def setImageItem(self, *args, **kwargs):
        self.slicer_area.sigViewOptionChanged.connect(self.limit_changed)

        self._span.blockSignals(True)
        self._span.setRegion(self.limits)
        self._span.blockSignals(False)
        self._span.sigRegionChanged.connect(self.level_change)
        self._span.sigRegionChangeFinished.connect(self.level_change_fin)
        self.color_changed()


class ItoolColorBar(pg.PlotWidget):
    def __init__(self, slicer_area: ImageSlicerArea | None = None, **cbar_kw):
        super().__init__(
            parent=slicer_area, plotItem=ItoolColorBarItem(slicer_area, **cbar_kw)
        )
        self.scene().sigMouseClicked.connect(self.mouseDragEvent)

    @property
    def cb(self) -> BetterColorBarItem:
        return self.plotItem

    def set_dimensions(
        self,
        width: int = 30,
        horiz_pad: int | None = None,
        vert_pad: int | None = None,
        font_size: float = 11.0,
    ):
        self.cb.set_dimensions(horiz_pad, vert_pad, font_size)
        self.setFixedWidth(width)

    def setVisible(self, visible: bool):
        super().setVisible(visible)
        self.cb.setVisible(visible)
        self.cb._span.blockSignals(not visible)

        if visible:
            self.cb._span.setRegion(self.cb.limits)



import numpy as np
import scipy.interpolate
from erlab.analysis.interpolate import interpn


def value_func_3d(x, y, z):
    return 2 * x + 3 * y - z


def test_interpn():
    x = np.linspace(0, 4, 5)
    y = np.linspace(0, 5, 6)
    z = np.linspace(0, 6, 7)
    points = (x, y, z)
    values = value_func_3d(*np.meshgrid(*points, indexing="ij"))
    point = np.array([[2.21, 3.12, 1.15], [2.67, 3.54, 1.03]])

    assert np.allclose(
        interpn(points, values, point),
        scipy.interpolate.interpn(
            points, values, point, method="linear", bounds_error=False
        ),
    )



import sys

import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtWidgets

from erlab.interactive.utilities import AnalysisWindow, ParameterGroup

__all__ = ["masktool"]


class PolyLineROIControls(ParameterGroup):
    def __init__(self, roi: pg.PolyLineROI, spinbox_kw=None, **kwargs):
        pass


class masktool(AnalysisWindow):
    # sigProgressUpdated = QtCore.Signal(int)

    def __init__(self, data, *args, **kwargs):
        super().__init__(
            data,
            *args,
            link="x",
            layout="horizontal",
            orientation="vertical",
            num_ax=2,
            data_is_input=False,
            **kwargs,
        )
        self._argnames = {}

        self.cursor = self.addParameterGroup(
            **{
                "Z dim": {
                    "qwtype": "combobox",
                    "items": self.data.dims,
                    "currentText": self.data.dims[-1],
                },
                "slider": {
                    "qwtype": "slider",
                    "orientation": QtCore.Qt.Horizontal,
                    "showlabel": False,
                    "value": 0,
                    "minimum": 0,
                    "maximum": self.data.shape[-1] - 1,
                },
                "Transpose": {"qwtype": "chkbox"},
            }
        )
        self.images[0].setDataArray(self.data[:, :, 0])
        self.cursor.sigParameterChanged.connect(self.update_cursor)

        self.roi = pg.PolyLineROI(
            [[0, 0]],
            parent=self.images[0],
            maxBounds=self.axes[0].getViewBox().itemBoundingRect(self.images[0]),
        )
        self.axes[0].addItem(self.roi)
        # self.roi.getArrayRegion(data, self.images[0], axes=(1,2))
        # self.roi.sigRegionChanged.connect()
        self.images[0]

    def update_cursor(self, change):
        # dim_z = change[-1]
        # update_only_values = True

        cursor_params = self.cursor.values

        dim_z = cursor_params["Z dim"]
        if "Z dim" in change:
            self.cursor.blockSignals(True)
            self.cursor.widgets["slider"].setMaximum(len(self.data[dim_z]) - 1)
            self.cursor.widgets["slider"].setValue(0)
            self.cursor.blockSignals(False)
            # update_only_values = False

        new_arr = self.data.isel({dim_z: self.cursor.widgets["slider"].value()})
        # if update_only_values:
        if cursor_params["Transpose"]:
            self.images[0].setDataArray(new_arr.T)
        else:
            self.images[0].setDataArray(new_arr)
        # else:
        # self.images[0].setDataArray(new_arr)
        # self.images[0].setImage(self.data.isel({dim_z:self.cursor.widgets["slider"].value()}).values)

        # self.cursor.values["slider"]


if __name__ == "__main__":
    import erlab.io

    qapp = QtWidgets.QApplication.instance()
    if not qapp:
        qapp = QtWidgets.QApplication(sys.argv)
    qapp.setStyle("Fusion")

    ds = erlab.io.load_igor_h5(
        "/Users/khan/Documents/ERLab/CsV3Sb5/220630_ALS_Kagome_nesting/maps.h5"
    )
    map3 = ds["Map3"].rename(phony_dim_0="kx", phony_dim_1="ky", phony_dim_2="eV")
    # map6 = ds["Map6"].rename(phony_dim_0="kx", phony_dim_3="ky", phony_dim_4="eV")
    ct = masktool(map3)
    ct.show()
    ct.activateWindow()
    ct.raise_()
    qapp.exec()



"""Interactive momentum conversion tool."""

__all__ = ["ktool"]

import os
import sys

import numpy as np
import numpy.typing as npt
import pyqtgraph as pg
import varname
import xarray as xr
from qtpy import QtCore, QtGui, QtWidgets, uic

import erlab.analysis
from erlab.interactive.colors import BetterColorBarItem  # noqa: F401
from erlab.interactive.colors import BetterImageItem
from erlab.interactive.colors import ColorMapComboBox  # noqa: F401
from erlab.interactive.colors import ColorMapGammaWidget  # noqa: F401
from erlab.interactive.imagetool import ImageTool
from erlab.interactive.utilities import array_rect, copy_to_clipboard, gen_function_code
from erlab.plotting.bz import get_bz_edge


class KspaceToolImageItem(BetterImageItem):
    def setDataArray(self, data=None, **kargs):
        rect = array_rect(data)
        if self.axisOrder == "row-major":
            img = np.ascontiguousarray(data.values)
        else:
            img = np.asfortranarray(data.values.T)
        pi = self.getPlotItem()
        if pi is not None:
            pi.setLabel("left", data.dims[0])
            pi.setLabel("bottom", data.dims[1])
        self.setImage(img, rect=rect, **kargs)

    def getPlotItem(self) -> pg.PlotItem | None:
        p = self
        while True:
            try:
                p = p.parentItem()
            except RuntimeError:
                return None
            if p is None:
                return None
            if isinstance(p, pg.PlotItem):
                return p


class KspaceToolGUI(
    *uic.loadUiType(os.path.join(os.path.dirname(__file__), "ktool.ui"))
):
    def __init__(self):
        # Start the QApplication if it doesn't exist
        self.qapp = QtCore.QCoreApplication.instance()
        if not self.qapp:
            self.qapp = QtWidgets.QApplication(sys.argv)
        self.qapp.setStyle("Fusion")

        # Initialize UI
        super().__init__()
        self.setupUi(self)
        self.setWindowTitle("Momentum Conversion")

        self.plotitems: tuple[pg.PlotItem, pg.PlotItem] = (pg.PlotItem(), pg.PlotItem())
        self.images: tuple[KspaceToolImageItem] = (
            KspaceToolImageItem(axisOrder="row-major"),
            KspaceToolImageItem(axisOrder="row-major"),
        )

        for i, plot in enumerate(self.plotitems):
            self.graphics_layout.addItem(plot, i, 0)
            plot.addItem(self.images[i])
            plot.showGrid(x=True, y=True, alpha=0.5)

        # Set up colormap controls
        self.cmap_combo.setDefaultCmap("terrain")
        self.cmap_combo.textActivated.connect(self.update_cmap)
        self.gamma_widget.setValue(0.5)
        self.gamma_widget.valueChanged.connect(self.update_cmap)
        self.invert_check.stateChanged.connect(self.update_cmap)
        self.invert_check.setChecked(True)
        self.contrast_check.stateChanged.connect(self.update_cmap)
        self.update_cmap()

        # Set up BZ controls
        self.bz_group.toggled.connect(self.update_bz)
        self.a_spin.valueChanged.connect(self.update_bz)
        self.b_spin.valueChanged.connect(self.update_bz)
        self.ang_spin.valueChanged.connect(self.update_bz)
        self.rot_spin.valueChanged.connect(self.update_bz)
        self.c_spin.valueChanged.connect(self.update_bz)
        self.ab_spin.valueChanged.connect(self.update_bz)
        self.n1_spin.valueChanged.connect(self.update_bz)
        self.n2_spin.valueChanged.connect(self.update_bz)
        self.reciprocal_check.stateChanged.connect(self.update_bz)
        self.points_check.stateChanged.connect(self.update_bz)

        self.plotitems[0].setVisible(False)
        self.angle_plot_check.stateChanged.connect(
            lambda: self.plotitems[0].setVisible(self.angle_plot_check.isChecked())
        )

    def update_cmap(self):
        name = self.cmap_combo.currentText()
        if name == self.cmap_combo.LOAD_ALL_TEXT:
            self.cmap_combo.load_all()
            return

        for im in self.images:
            im.set_colormap(
                name,
                self.gamma_widget.value(),
                reverse=self.invert_check.isChecked(),
                highContrast=self.contrast_check.isChecked(),
                update=True,
            )

    def get_bz_lines(self):
        raise NotImplementedError

    def update_bz(self):
        self.plotitems[1].clearPlots()
        if not self.bz_group.isChecked():
            return
        if self.reciprocal_check.isChecked():
            self.a_spin.setSuffix(" Å⁻¹")
            self.b_spin.setSuffix(" Å⁻¹")
            self.c_spin.setSuffix(" Å⁻¹")
            self.ab_spin.setSuffix(" Å⁻¹")
        else:
            self.a_spin.setSuffix(" Å")
            self.b_spin.setSuffix(" Å")
            self.c_spin.setSuffix(" Å")
            self.ab_spin.setSuffix(" Å")

        lines, vertices = self.get_bz_lines()
        for line in lines:
            self.plotitems[1].plot(line[:, 0], line[:, 1], pen=pg.mkPen("m", width=2))
            vertices = np.vstack((vertices, np.mean(line, axis=0)))

        if self.points_check.isChecked():
            self.plotitems[1].plot(
                vertices[:, 0],
                vertices[:, 1],
                symbol="o",
                pen=pg.mkColor(255, 255, 255, 0),
                symbolPen=pg.mkColor(255, 255, 255, 0),
                symbolBrush=pg.mkColor("m"),
                symbolSize=6,
            )

    def __post_init__(self, execute=None):
        self.show()
        self.raise_()
        self.activateWindow()

        if execute is None:
            execute = True
            try:
                shell = get_ipython().__class__.__name__  # type: ignore
                if shell in ["ZMQInteractiveShell", "TerminalInteractiveShell"]:
                    execute = False
                    from IPython.lib.guisupport import start_event_loop_qt4

                    start_event_loop_qt4(self.qapp)
            except NameError:
                pass
        if execute:
            self.qapp.exec()


class KspaceTool(KspaceToolGUI):
    def __init__(self, data: xr.DataArray, *, data_name: str | None = None):
        super().__init__()

        self._argnames = {}

        if data_name is None:
            try:
                self._argnames["data"] = varname.argname(
                    "data", func=self.__init__, vars_only=False
                )
            except varname.VarnameRetrievingError:
                self._argnames["data"] = "data"
        else:
            self._argnames["data"] = data_name

        self.data: xr.DataArray = data.copy(deep=True)

        if self.data.kspace.has_eV:
            self.center_spin.setRange(self.data.eV[0], self.data.eV[-1])
            self.width_spin.setRange(1, len(self.data.eV))
            self.center_spin.valueChanged.connect(self.update)
            self.width_spin.valueChanged.connect(self.update)
        else:
            self.energy_group.setDisabled(True)

        self.bounds_group.toggled.connect(self.update)
        self.resolution_supergroup.toggled.connect(self.update)

        self._offset_spins: dict[str, QtWidgets.QDoubleSpinBox] = {}
        offset_labels = {"delta": "𝛿", "chi": "𝜒₀", "xi": "𝜉₀", "beta": "𝛽₀"}
        for k in self.data.kspace.valid_offset_keys:
            self._offset_spins[k] = QtWidgets.QDoubleSpinBox()
            self._offset_spins[k].setRange(-180, 180)
            self._offset_spins[k].setSingleStep(0.01)
            self._offset_spins[k].setDecimals(3)
            self._offset_spins[k].setValue(self.data.kspace.offsets[k])
            self._offset_spins[k].valueChanged.connect(self.update)
            self._offset_spins[k].setSuffix("°")
            self.offsets_group.layout().addRow(offset_labels[k], self._offset_spins[k])

        if self.data.kspace.has_hv:
            self._offset_spins["V0"] = QtWidgets.QDoubleSpinBox()
            self._offset_spins["V0"].setRange(0, 100)
            self._offset_spins["V0"].setSingleStep(1)
            self._offset_spins["V0"].setDecimals(1)
            self._offset_spins["V0"].setValue(self.data.kspace.inner_potential)
            self._offset_spins["V0"].valueChanged.connect(self.update)
            self._offset_spins["V0"].setSuffix(" eV")
            self.offsets_group.layout().addRow("V₀", self._offset_spins["V0"])

            for i in range(8):
                self.bz_form.setRowVisible(i, i not in (0, 1, 2))
        else:
            for i in range(8):
                self.bz_form.setRowVisible(i, i not in (3, 4))

        self._bound_spins: dict[str, QtWidgets.QDoubleSpinBox] = {}
        self._resolution_spins: dict[str, QtWidgets.QDoubleSpinBox] = {}
        bounds = self.data.kspace.estimate_bounds()
        for k in self.data.kspace.momentum_axes:
            for j in range(2):
                name = f"{k}{j}"
                self._bound_spins[name] = QtWidgets.QDoubleSpinBox()
                if k == "kz":
                    self._bound_spins[name].setRange(0, 100)
                else:
                    self._bound_spins[name].setRange(-10, 10)
                self._bound_spins[name].setSingleStep(0.01)
                self._bound_spins[name].setDecimals(2)
                self._bound_spins[name].setValue(bounds[k][j])
                self._bound_spins[name].valueChanged.connect(self.update)
                self._bound_spins[name].setSuffix(" Å⁻¹")
                self.bounds_group.layout().addRow(name, self._bound_spins[name])

            self._resolution_spins[k] = QtWidgets.QDoubleSpinBox()
            self._resolution_spins[k].setRange(0.001, 10)
            self._resolution_spins[k].setSingleStep(0.001)
            self._resolution_spins[k].setDecimals(4)
            self._resolution_spins[k].setValue(self.data.kspace.estimate_resolution(k))
            self._resolution_spins[k].valueChanged.connect(self.update)
            self._resolution_spins[k].setSuffix(" Å⁻¹")
            self.resolution_group.layout().addRow(k, self._resolution_spins[k])

        self.res_btn.clicked.connect(self.calculate_resolution)
        self.res_npts_check.toggled.connect(self.calculate_resolution)

        for pi in self.plotitems:
            if self.data.kspace.has_beta and not self.data.kspace.has_hv:
                pi.vb.setAspectLocked(lock=True, ratio=1)
        self.open_btn.clicked.connect(self.show_converted)
        self.copy_btn.clicked.connect(self.copy_code)
        self.update()

        self.__post_init__()

    def calculate_resolution(self):
        for k, spin in self._resolution_spins.items():
            spin.setValue(
                self.data.kspace.estimate_resolution(
                    k, from_numpoints=self.res_npts_check.isChecked()
                )
            )

    def show_converted(self):
        self.data.kspace.offsets = self.offset_dict

        if self.data.kspace.has_hv:
            self.data.kspace.inner_potential = self._offset_spins["V0"].value()

        wait_dialog = QtWidgets.QDialog(self)
        wait_dialog.setLayout(QtWidgets.QVBoxLayout())
        wait_dialog.layout().addWidget(QtWidgets.QLabel("Converting..."))
        wait_dialog.open()
        itool = ImageTool(
            self.data.kspace.convert(bounds=self.bounds, resolution=self.resolution)
        )
        wait_dialog.close()
        itool.show()

    def copy_code(self):
        arg_dict = {}
        if self.bounds is not None:
            arg_dict["bounds"] = self.bounds
        if self.resolution is not None:
            arg_dict["resolution"] = self.resolution

        # Detected input name must be single identifier.
        # Otherwise the generated code will not apply offsets correctly.
        input_name: str = str(self._argnames["data"])
        if not input_name.isidentifier():
            input_name = "data"

        out_lines: list[str] = []

        if self.data.kspace.has_hv:
            out_lines.append(
                f"{input_name}.kspace.inner_potential"
                f" = {self._offset_spins['V0'].value()}"
            )
        out_lines.append(f"{input_name}.kspace.offsets = {self.offset_dict}")
        out_lines.append(
            gen_function_code(
                copy=False,
                **{f"{input_name}_kconv = {input_name}.kspace.convert": [arg_dict]},
            )
        )

        copy_to_clipboard(out_lines)

    @property
    def bounds(self) -> dict[str, tuple[float, float]] | None:
        if self.bounds_group.isChecked():
            return {
                k: tuple(self._bound_spins[f"{k}{j}"].value() for j in range(2))
                for k in self.data.kspace.momentum_axes
            }
        else:
            return None

    @property
    def resolution(self) -> dict[str, float] | None:
        if self.resolution_supergroup.isChecked():
            return {
                k: self._resolution_spins[k].value()
                for k in self.data.kspace.momentum_axes
            }
        else:
            return None

    @property
    def offset_dict(self) -> dict[str, float]:
        return {
            k: np.round(self._offset_spins[k].value(), 5)
            for k in self.data.kspace.valid_offset_keys
        }

    def _angle_data(self) -> xr.DataArray:
        if self.data.kspace.has_eV:
            center, width = self.center_spin.value(), self.width_spin.value()
            if width == 0:
                return self.data.sel(eV=center, method="nearest")
            else:
                arr = self.data.eV.values
                idx = np.searchsorted((arr[:-1] + arr[1:]) / 2, center)
                return (
                    self.data.isel(
                        eV=slice(idx - width // 2, idx + (width - 1) // 2 + 1)
                    )
                    .mean("eV", skipna=True, keep_attrs=True)
                    .assign_coords(eV=center)
                )
        else:
            return self.data

    def get_data(self) -> tuple[xr.DataArray, xr.DataArray]:
        # Set angle offsets
        data_ang = self._angle_data()
        data_ang.kspace.offsets = self.offset_dict

        if self.data.kspace.has_hv:
            data_ang.kspace.inner_potential = self._offset_spins["V0"].value()

        # Convert to kspace
        data_k = data_ang.kspace.convert(
            bounds=self.bounds, resolution=self.resolution, silent=True
        )
        return data_ang, data_k

    def update(self):
        ang, k = self.get_data()
        self.images[0].setDataArray(ang.T)
        self.images[1].setDataArray(k.T)

    def get_bz_lines(
        self,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        if self.data.kspace.has_hv:
            # Out-of-plane BZ
            a, c = self.ab_spin.value(), self.c_spin.value()
            rot = np.deg2rad(self.rot_spin.value())

            basis = np.array([[a, 0], [0, c]])
            if not self.reciprocal_check.isChecked():
                basis = 2 * np.pi * np.linalg.inv(basis).T

            if rot != 0.0:
                basis[0, :] *= np.cos(rot)

            lines, vertices = get_bz_edge(
                basis,
                reciprocal=True,
                extend=(self.n1_spin.value(), self.n2_spin.value()),
            )

        else:
            # In-plane BZ
            a, b = self.a_spin.value(), self.b_spin.value()
            ang = np.deg2rad(self.ang_spin.value())
            rot = np.deg2rad(self.rot_spin.value())

            avec = np.array([[a, 0], [b * np.cos(ang), b * np.sin(ang)]])

            lines, vertices = get_bz_edge(
                avec,
                reciprocal=self.reciprocal_check.isChecked(),
                extend=(self.n1_spin.value(), self.n2_spin.value()),
            )

            if rot != 0.0:
                rotmat = np.array(
                    [[np.cos(rot), -np.sin(rot)], [np.sin(rot), np.cos(rot)]]
                )
                lines = (rotmat @ lines.transpose(1, 2, 0)).transpose(2, 0, 1)
                vertices = (rotmat @ vertices.T).T

        return lines, vertices

    def closeEvent(self, event: QtGui.QCloseEvent):
        del self.data
        super().closeEvent(event)


def ktool(data: xr.DataArray, *, data_name: str | None = None) -> KspaceTool:
    """Interactive momentum conversion tool."""
    if data_name is None:
        try:
            data_name = varname.argname("data", func=ktool, vars_only=False)
        except varname.VarnameRetrievingError:
            data_name = "data"
    return KspaceTool(data, data_name=data_name)


if __name__ == "__main__":
    dat = erlab.io.load_hdf5("/Users/khan/2210_ALS_f0008.h5")
    win = ktool(dat)



"""Utilities for plotting Brillouin zones."""

__all__ = ["get_bz_edge", "plot_hex_bz"]

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import itertools
import scipy.spatial
from matplotlib.patches import RegularPolygon

from erlab.plotting.colors import axes_textcolor

abbrv_kws = {
    "facecolor": ["fc", "none"],
    "linestyle": ["ls", "--"],
    "linewidth": ["lw", 0.5],
}


def get_bz_edge(
    basis: npt.NDArray[np.float64],
    reciprocal: bool = True,
    extend: tuple[int, int, int] | tuple[int, int] | None = None,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """Calculates the edge of the first Brillouin zone (BZ) from lattice vectors.

    Parameters
    ----------
    basis
        ``(N, N)`` numpy array where ``N = 2``or ``3`` with each row containing the
        lattice vectors.
    reciprocal
        If `False`, the `basis` are given in real space lattice vectors.
    extend
        Tuple of positive integers specifying the number of times to extend the BZ in
        each direction. If `None`, only the first BZ is returned (equivalent to ``(1,) *
        N``).

    Returns
    -------
    lines : array-like
        ``(M, 2, N)`` array that specifies the endpoints of the ``M`` lines that make up
        the BZ edge, where ``N = len(basis)``.
    vertices : array-like
        Vertices of the BZ.

    """
    if not (basis.shape == (2, 2) or basis.shape == (3, 3)):
        raise ValueError("Shape of `basis` must be (N, N) where N = 2 or 3.")
    if not reciprocal:
        basis = 2 * np.pi * np.linalg.inv(basis).T

    ndim = basis.shape[-1]

    if extend is None:
        extend = (1,) * ndim

    points = (
        np.tensordot(basis, np.mgrid[[slice(-1, 2) for _ in range(ndim)]], axes=[0, 0])
        .reshape((ndim, 3**ndim))
        .T
    )

    # Get index of origin
    zero_ind = np.where((points == 0).all(axis=1))[0][0]

    vor = scipy.spatial.Voronoi(points)

    lines = []
    vertices = []

    for pointidx, simplex in zip(vor.ridge_points, vor.ridge_vertices):
        simplex = np.asarray(simplex)
        if zero_ind in pointidx:
            # If the origin is included in the ridge, add the vertices
            lines.append(vor.vertices[np.r_[simplex, simplex[0]]])
            vertices.append(vor.vertices[simplex])

    # Remove duplicates
    lines_new = []
    vertices_new = []

    for line in lines:
        for i in range(line.shape[0] - 1):
            if not any(
                np.allclose(line[i : i + 2], line_new)
                or np.allclose(line[i : i + 2], np.flipud(line_new))
                for line_new in lines_new
            ):
                lines_new.append(line[i : i + 2])
    for v in np.r_[*vertices]:
        if not any(np.allclose(v, vn) for vn in vertices_new):
            vertices_new.append(v)

    lines = np.asarray(lines_new)
    vertices = np.asarray(vertices_new)

    # Extend the BZ
    additional_lines = []
    additional_verts = []
    for vals in itertools.product(*[range(-n + 1, n) for n in extend]):
        if vals != (0,) * ndim:
            displacement = np.dot(vals, basis)
            additional_lines.append(lines + displacement)
            additional_verts.append(vertices + displacement)
    lines = np.r_[lines, *additional_lines]
    vertices = np.r_[vertices, *additional_verts]

    return lines, vertices


def plot_hex_bz(
    a=3.54, rotate=0.0, offset=(0.0, 0.0), reciprocal=True, ax=None, **kwargs
):
    """
    Plots a 2D hexagonal BZ overlay on the specified axes.
    """
    kwargs.setdefault("zorder", 5)
    for k, v in abbrv_kws.items():
        kwargs[k] = kwargs.pop(k, kwargs.pop(*v))
    if ax is None:
        ax = plt.gca()
    if np.iterable(ax):
        return [
            plot_hex_bz(a=a, rotate=rotate, offset=offset, ax=x, **kwargs) for x in ax
        ]
    else:
        kwargs["edgecolor"] = kwargs.pop(
            "edgecolor", kwargs.pop("ec", axes_textcolor(ax))
        )
    if reciprocal:
        r = 4 * np.pi / (a * 3)
    else:
        r = 2 * a
    clip = kwargs.pop("clip_path", None)
    poly = RegularPolygon(offset, 6, radius=r, orientation=np.deg2rad(rotate), **kwargs)
    ax.add_patch(poly)
    if clip is not None:
        poly.set_clip_path(clip)
    return poly



"""Functions related to analyzing temperature-dependent resistance data.

Currently only supports loading raw data from ``.dat`` and ``.csv`` files output by
physics lab III equipment.

"""

import re
import os
from io import StringIO

import numpy as np
import xarray as xr
import pandas as pd

__all__ = ["load_resistance_physlab"]


def load_resistance_physlab(path: str, **kwargs) -> xr.Dataset:
    """Loads resistance measurement acquired with physics lab III equipment.

    Parameters
    ----------
    path
        Local path to ``.dat`` file.

    Returns
    -------
    ds : xarray.Dataset
        Dataset containing resistance data from the file.

    """
    if os.path.splitext(path)[1] == ".dat":
        return _load_resistance_physlab_old(path, **kwargs)
    else:
        return (
            pd.read_csv(path, index_col=0, usecols=[1, 2, 3, 4])
            .to_xarray()
            .rename(
                {
                    "Elapsed Time (s)": "time",
                    "Temperature (K)": "temp",
                    "Resistance (Ohm)": "res",
                    "Current (A)": "curr",
                }
            )
        )


def _load_resistance_physlab_old(
    path: str, encoding: str = "windows-1252", **kwargs
) -> xr.Dataset:
    """Loads resistance measurement acquired with physics lab III equipment.

    Parameters
    ----------
    path
        Local path to ``.dat`` file.
    encoding
        Encoding passed onto `open`, defaults to ``'windows-1252'``.

    Returns
    -------
    ds : xarray.Dataset
        Dataset containing resistance data from the file.

    Examples
    --------
    Load from file:

    >>> data = load_resistance_physlab("/path/to/example_data.dat")
    >>> data
    <xarray.Dataset>
    Dimensions:  (temp: 1087)
    Coordinates:
    * temp     (temp) float64 21.44 21.61 21.81 ... 299.4 299.6 299.8
    Data variables:
        time     (temp) float64 44.45 66.06 ... 2.346e+04 2.348e+04
        res      (temp) float64 0.02837 0.02886 ... 0.05812 0.05914
        curr     (temp) float64 0.0001 0.0001 ... 0.0001 0.0001
        temperr  (temp) float64 0.098 0.11 ... 0.066 0.046
        reserr   (temp) float64 0.001551 0.001202 ... 0.001027 0.001125

    Plot resistance versus temperature:

    >>> data.res.plot()

    """
    with open(path, encoding=encoding) as file:
        content = re.sub(
            r"(e[-+]\d{3}) {2,3}(-?)",
            "\\g<1>\\t \\g<2>",
            file.read(),
            count=0,
            flags=re.MULTILINE,
        )
    content = content.replace("-1.#IO", "   nan")
    data = np.genfromtxt(
        StringIO(content),
        delimiter="\t",
        skip_header=3,
        usecols=[2, 3, 4, 5, 6, 7],
        **kwargs,
    )
    ds = xr.Dataset(
        data_vars={
            "time": ("temp", data[:, 0]),
            "res": ("temp", data[:, 2]),
            "curr": ("temp", data[:, 3]),
            "temp_err": ("temp", data[:, 4]),
            "res_err": ("temp", data[:, 5]),
        },
        coords={
            "temp": data[:, 1],
        },
    )
    return ds



"""Fermi edge fitting."""

__all__ = [
    "edge",
    "poly",
    "poly_from_edge",
    "spline_from_edge",
    "resolution",
    "resolution_roi",
]

from collections.abc import Sequence

import arpes
import arpes.fits
import lmfit.model
import matplotlib
import matplotlib.figure
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.interpolate
import uncertainties
import xarray as xr

from erlab.analysis.fit.models import (
    ExtendedAffineBroadenedFD,
    PolynomialModel,
    StepEdgeModel,
)
from erlab.analysis.utilities import correct_with_edge
from erlab.plotting.general import autoscale_to, figwh


def edge(
    gold: xr.DataArray | xr.Dataset,
    angle_range: tuple[float, float],
    eV_range: tuple[float, float],
    bin_size: tuple[int, int] = (1, 1),
    temp: float | None = None,
    vary_temp: bool = False,
    fast: bool = False,
    method: str = "leastsq",
    progress: bool = True,
    parallel_kw: dict | None = None,
    return_full: bool = False,
    fixed_center: float | None = None,
    scale_covar: bool = True,
    **kwargs: dict,
) -> tuple[npt.NDArray, npt.NDArray] | xr.Dataset:
    """
    Fit a Fermi edge to the given gold data.

    Parameters
    ----------
    gold
        The gold data to fit the edge model to.
    angle_range
        The range of alpha values to consider.
    eV_range
        The range of eV values to consider.
    bin_size
        The bin size for coarsening the gold data, by default (1, 1).
    temp
        The temperature in Kelvins. If `None`, the temperature is inferred from the
        attributes, by default `None`
    vary_temp
        Whether to fit the temperature value during fitting, by default `False`.
    fast
        Whether to use the Gaussian-broadeded step function to fit the edge, by default
        `False`.
    method
        The fitting method to use, by default ``"leastsq"``.
    progress
        Whether to display the fitting progress, by default `True`.
    parallel_kw
        Additional keyword arguments for parallel fitting, by default `None`.
    return_full
        Whether to return the full fit results, by default `False`.
    fixed_center
        The fixed center value. If provided, the Fermi level will be fixed at the given
        value, by default `None`.
    scale_covar
        Whether to scale the covariance matrix, by default `True`.
    **kwargs
        Additional keyword arguments to fitting.

    Returns
    -------
    center_arr, center_stderr
        The fitted center values and their standard errors, returned when `return_full`
        is `False`.
    fitresults
        A dataset containing the full fit results, returned when `return_full` is
        `True`.

    """
    if fast:
        params = {}
        model_cls = StepEdgeModel
    else:
        if temp is None:
            temp = gold.attrs["temp_sample"]
        params = {
            "temp": {"value": temp, "vary": vary_temp},
        }
        model_cls = ExtendedAffineBroadenedFD

    if parallel_kw is None:
        parallel_kw = {}

    if fixed_center is not None:
        params["center"] = {"value": fixed_center, "vary": False}

    if any(b != 1 for b in bin_size):
        gold_binned = gold.coarsen(alpha=bin_size[0], eV=bin_size[1], boundary="trim")
        gold = gold_binned.mean()

    gold_sel = gold.sel(alpha=slice(*angle_range), eV=slice(*eV_range))

    # Assuming Poisson noise, the weights are the square root of the counts.
    weights = np.sqrt(gold.sum("eV"))

    fitresults = arpes.fits.broadcast_model(
        model_cls,
        gold_sel,
        "alpha",
        weights=weights,
        params=params,
        method=method,
        progress=progress,
        parallel_kw=parallel_kw,
        scale_covar=scale_covar,
        **kwargs,
    )
    if return_full:
        return fitresults

    center_arr, center_stderr = fitresults.F.p("center"), fitresults.F.s("center")
    center_arr = center_arr.where(~center_stderr.isnull(), drop=True)
    center_stderr = center_stderr.where(~center_stderr.isnull(), drop=True)
    return center_arr, center_stderr


def poly_from_edge(
    center, weights=None, degree=4, method="leastsq", scale_covar=True
) -> lmfit.model.ModelResult:
    modelresult = PolynomialModel(degree=degree).guess_fit(
        center, weights=weights, method=method, scale_covar=scale_covar
    )
    return modelresult


def spline_from_edge(
    center, weights: Sequence[float] | None = None, lam: float | None = None
) -> scipy.interpolate.BSpline:
    spl = scipy.interpolate.make_smoothing_spline(
        center.alpha.values,
        center.values,
        w=np.asarray(weights),
        lam=lam,
    )
    return spl


def _plot_gold_fit(fig, gold, angle_range, eV_range, center_arr, center_stderr, res):
    if isinstance(res, lmfit.model.ModelResult):
        is_callable = False
    elif callable(res):
        is_callable = True
    else:
        raise ValueError

    if not isinstance(fig, plt.Figure):
        fig = plt.figure(figsize=figwh(0.75, wscale=1.75))

    gs = fig.add_gridspec(2, 2, height_ratios=[1, 3])
    ax0 = fig.add_subplot(gs[:, 0])
    ax1 = fig.add_subplot(gs[0, 1])
    plt.tick_params("x", labelbottom=False)
    ax2 = fig.add_subplot(gs[1, 1], sharex=ax1)

    gold.qplot(ax=ax0, cmap="copper", gamma=0.5)
    rect = mpatches.Rectangle(
        (angle_range[0], eV_range[0]),
        np.diff(angle_range)[0],
        np.diff(eV_range)[0],
        ec="w",
        alpha=0.5,
        lw=0.75,
        fc="none",
    )
    ax0.add_patch(rect)
    ax0.errorbar(
        center_arr.alpha,
        center_arr,
        center_stderr,
        fmt="o",
        lw=0.5,
        mfc="w",
        zorder=0,
        ms=2,
    )

    if is_callable:
        ax0.plot(gold.alpha, res(gold.alpha), "r-", lw=0.75)
    else:
        ax0.plot(gold.alpha, res.eval(res.params, x=gold.alpha), "r-", lw=0.75)
    ax0.set_ylim(gold.eV[[0, -1]])

    data_kws = {"lw": 0.5, "ms": 2, "mfc": "w", "zorder": 0, "c": "0.4", "capsize": 0}
    fit_kws = {"c": "r", "lw": 0.75}

    if is_callable:
        residuals = res(center_arr.alpha.values) - center_arr.values
        x_eval = np.linspace(
            min(center_arr.alpha.values),
            max(center_arr.alpha.values),
            3 * len(center_arr.alpha),
        )
        ax1.axhline(0, **fit_kws)
        ax1.errorbar(
            center_arr.alpha,
            residuals,
            yerr=lmfit.model.propagate_err(
                center_arr.values, center_stderr.values, "abs"
            ),
            fmt="o",
            **data_kws,
        )
        ax1.set_ylabel("residuals")

        ax2.errorbar(
            center_arr.alpha,
            center_arr.values,
            yerr=lmfit.model.propagate_err(
                center_arr.values, center_stderr.values, "abs"
            ),
            fmt="o",
            label="data",
            **data_kws,
        )
        ax2.plot(x_eval, res(x_eval), "-", label="best fit", **fit_kws)
        ax2.legend()
        ax1.set_ylim(autoscale_to(residuals))
        ax2.set_ylim(autoscale_to(center_arr.values))
    else:
        res.plot_residuals(
            ax=ax1,
            data_kws=data_kws,
            fit_kws=fit_kws,
        )
        res.plot_fit(
            ax=ax2,
            data_kws=data_kws,
            fit_kws=fit_kws,
            numpoints=3 * len(center_arr.alpha),
        )
        ax1.set_ylim(autoscale_to(res.eval() - res.data))
        ax2.set_ylim(autoscale_to(res.data))

    ax1.set_title("")
    ax2.set_title("")


def poly(
    gold: xr.DataArray | xr.Dataset,
    angle_range: tuple[float, float],
    eV_range: tuple[float, float],
    bin_size: tuple[int, int] = (1, 1),
    temp: float | None = None,
    vary_temp: bool = False,
    fast: bool = False,
    method: str = "leastsq",
    degree: int = 4,
    correct: bool = False,
    crop_correct: bool = False,
    parallel_kw: dict | None = None,
    plot: bool = True,
    fig: matplotlib.figure.Figure | None = None,
    scale_covar: bool = True,
    scale_covar_edge: bool = True,
) -> lmfit.model.ModelResult | tuple[lmfit.model.ModelResult, xr.DataArray]:
    center_arr, center_stderr = edge(
        gold,
        angle_range,
        eV_range,
        bin_size=bin_size,
        temp=temp,
        vary_temp=vary_temp,
        fast=fast,
        method=method,
        parallel_kw=parallel_kw,
        scale_covar=scale_covar_edge,
    )

    modelresult = poly_from_edge(
        center_arr,
        weights=1 / center_stderr,
        degree=degree,
        method=method,
        scale_covar=scale_covar,
    )
    if plot:
        _plot_gold_fit(
            fig, gold, angle_range, eV_range, center_arr, center_stderr, modelresult
        )
    if correct:
        if crop_correct:
            gold = gold.sel(alpha=slice(*angle_range), eV=slice(*eV_range))
        corr = correct_with_edge(gold, modelresult, plot=False)
        return modelresult, corr
    else:
        return modelresult


def spline(
    gold: xr.DataArray | xr.Dataset,
    angle_range: tuple[float, float],
    eV_range: tuple[float, float],
    bin_size: tuple[int, int] = (1, 1),
    temp: float | None = None,
    vary_temp: bool = False,
    fast: bool = False,
    method: str = "leastsq",
    lam: float | None = None,
    correct: bool = False,
    crop_correct: bool = False,
    parallel_kw: dict | None = None,
    plot: bool = True,
    fig: matplotlib.figure.Figure | None = None,
    scale_covar_edge: bool = True,
) -> scipy.interpolate.BSpline | tuple[scipy.interpolate.BSpline, xr.DataArray]:
    center_arr, center_stderr = edge(
        gold,
        angle_range,
        eV_range,
        bin_size=bin_size,
        temp=temp,
        vary_temp=vary_temp,
        fast=fast,
        method=method,
        parallel_kw=parallel_kw,
        scale_covar=scale_covar_edge,
    )

    spl = spline_from_edge(center_arr, weights=1 / center_stderr, lam=lam)
    if plot:
        _plot_gold_fit(fig, gold, angle_range, eV_range, center_arr, center_stderr, spl)
    if correct:
        if crop_correct:
            gold = gold.sel(alpha=slice(*angle_range), eV=slice(*eV_range))
        corr = correct_with_edge(gold, spl, plot=False)
        return spl, corr
    else:
        return spl


def resolution(
    gold: xr.DataArray | xr.Dataset,
    angle_range: tuple[float, float],
    eV_range_edge: tuple[float, float],
    eV_range_fit: tuple[float, float] | None = None,
    bin_size: tuple[int, int] = (1, 1),
    degree: int = 4,
    fast: bool = False,
    method: str = "leastsq",
    plot: bool = True,
    parallel_kw: dict | None = None,
    scale_covar: bool = True,
) -> lmfit.model.ModelResult:
    pol, gold_corr = poly(
        gold,
        angle_range,
        eV_range_edge,
        bin_size=bin_size,
        degree=degree,
        correct=True,
        fast=fast,
        method=method,
        plot=plot,
        parallel_kw=parallel_kw,
    )

    if eV_range_fit is None:
        eV_range_fit = tuple(r - np.mean(pol.best_fit) for r in eV_range_edge)
    del pol
    gold_roi = gold_corr.sel(alpha=slice(*angle_range))
    edc_avg = gold_roi.mean("alpha").sel(eV=slice(*eV_range_fit))

    params = {
        "temp": {"value": gold.attrs["temp_sample"], "vary": False},
        "resolution": {"value": 0.1, "vary": True, "min": 0},
    }
    fit = ExtendedAffineBroadenedFD().guess_fit(
        edc_avg, params=params, method=method, scale_covar=scale_covar
    )
    if plot:
        plt.show()
        ax = plt.gca()
        gold_corr.qplot(ax=ax, cmap="copper", gamma=0.5)
        rect = mpatches.Rectangle(
            (angle_range[0], eV_range_fit[0]),
            np.diff(angle_range)[0],
            np.diff(eV_range_fit)[0],
            ec="w",
            alpha=0.5,
            lw=0.75,
            fc="none",
        )
        ax.add_patch(rect)
        ax.set_ylim(gold_corr.eV[[0, -1]])

        fit.plot(
            data_kws={"lw": 0.75, "ms": 4, "mfc": "w", "zorder": 0, "c": "0.4"},
            fit_kws={"c": "r", "lw": 1.5},
        )

    center_uf = uncertainties.ufloat(fit.params["center"], fit.params["center"].stderr)
    res_uf = uncertainties.ufloat(
        fit.params["resolution"], fit.params["resolution"].stderr
    )
    print(f"center = {center_uf:S} eV\nresolution = {res_uf:.4S} eV")
    return fit


def resolution_roi(
    gold_roi: xr.DataArray,
    eV_range: tuple[float, float],
    fix_temperature: bool = True,
    method: str = "leastsq",
    plot: bool = True,
    scale_covar: bool = True,
) -> lmfit.model.ModelResult:
    edc_avg = gold_roi.mean("alpha").sel(eV=slice(*eV_range))

    params = {
        "temp": {"value": gold_roi.attrs["temp_sample"], "vary": not fix_temperature},
        "resolution": {"value": 0.1, "vary": True, "min": 0},
    }
    fit = ExtendedAffineBroadenedFD().guess_fit(
        edc_avg,
        params=params,
        method=method,
        scale_covar=scale_covar,  # weights=1 / edc_stderr
    )
    if plot:
        ax = plt.gca()
        gold_roi.qplot(ax=ax, cmap="copper", gamma=0.5)
        ax.fill_between(
            gold_roi.alpha,
            eV_range[0],
            eV_range[1],
            ec="w",
            fc="none",
            alpha=0.4,
            lw=0.75,
        )
        ax.set_ylim(gold_roi.eV[[0, -1]])

        fit.plot(
            data_kws={"lw": 0.75, "ms": 4, "mfc": "w", "zorder": 0, "c": "0.4"},
            fit_kws={"c": "r", "lw": 1.5},
        )

    center_uf = uncertainties.ufloat(fit.params["center"], fit.params["center"].stderr)
    res_uf = uncertainties.ufloat(
        fit.params["resolution"], fit.params["resolution"].stderr
    )
    print(f"center = {center_uf:S} eV\nresolution = {res_uf:.4S} eV")
    return fit
